# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2013       Oleh Petriv <olehsbox[at]yahoo.com>
# Copyright (C) 2013       Fedir Zinchuk <fedikw[at]gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gramps/plugins/rel/rel_uk.py
# UA: Пояснення щодо родинних відносин див. relationship.py
# EN: Ukrainian relationship calculator. For more information see relationship.py

"""
Ukrainian-specific definitions of relationships
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------


# Relationship levels

_level_name = [
    "",
    "рідний(-на)",
    "двоюрідний(-на)",
    "троюрідний(-на)",
    "чотириюрідний(-на)",
    "п'ятиюрідний(-на)",
    "шестиюрідний(-на)",
    "семиюрідний(-на)",
    "восьмиюрідний(-на)",
    "дев'ятиюрідний(-на)",
    "десятиюрідний(-на)",
    "одинадцятиюрідний(-на)",
    "дванадцятиюрідний(-на)",
    "тринадцятиюрідний(-на)",
    "чотирнадцятиюрідний(-на)",
    "п'ятнадцятиюрідний(-на)",
    "шістнадцятиюрідний(-на)",
    "сімнадцятиюрідний(-на)",
    "вісімнадцятиюрідний(-на)",
    "дев'ятнадцятиюрідний(-на)",
    "двадцятиюрідний(-на)",
]
_level_name_male = [
    "",
    "рідний",
    "двоюрідний",
    "троюрідний",
    "чотириюрідний",
    "п'ятиюрідний",
    "шестиюрідний",
    "семиюрідний",
    "восьмиюрідний",
    "дев'ятиюрідний",
    "десятиюрідний",
    "одинадцятиюрідний",
    "дванадцятиюрідний",
    "тринадцятиюрідний",
    "чотирнадцятиюрідний",
    "п'ятнадцятиюрідний",
    "шістнадцятиюрідний",
    "сімнадцятиюрідний",
    "вісімнадцятиюрідний",
    "дев'ятнадцятиюрідний",
    "двадцятиюрідний",
]
_level_name_female = [
    "",
    "рідна",
    "двоюрідна",
    "троюрідна",
    "чотириюріднa",
    "п'ятиюрідна",
    "шестиюрідна",
    "семиюрідна",
    "восьмиюрідна",
    "дев'ятиюрідна",
    "десятиюрідна",
    "одинадцятиюрідна",
    "дванадцятиюрідна",
    "тринадцятиюрідна",
    "чотирнадцятиюрідна",
    "п'ятнадцятиюрідна",
    "шістнадцятиюрідна",
    "сімнадцятиюрідна",
    "вісімнадцятиюрідна",
    "дев'ятнадцятиюрідна",
    "двадцятиюрідна",
]

_son_level = [
    "",
    "син",
    "внук",
    "правнук",
    "праправнук",
]

_daughter_level = [
    "",
    "дочка",
    "онука",
    "правнучка",
    "праправнучка",
]

_father_level = [
    "",
    "батько",
    "дід",
    "прадід",
    "прапрадід",
]

_mother_level = [
    "",
    "мати",
    "баба",
    "прабаба",
    "прапрабаба",
]

_uncle_level = [
    "",
    "брат",  # not used, just for keep count
    "дядько",
    "дід",
    "прадід",
    "прапрадід",
]

_aunt_level = [
    "",
    "сестра",  # not used, just for keep count
    "тітка",
    "баба",
    "прабаба",
    "прапрабаба",
]


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------


class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_sword_distaff(self, level, reltocommon, gender=Person.UNKNOWN):
        """
        Generate relationships 'by male line' or 'by female line',
        specific for Ukraine
        """
        if level <= 1:
            return ""

        # test the relation line
        by_line = ""
        for g in reltocommon:
            if by_line and by_line != g:
                by_line = ""
                break
            by_line = g

        if by_line == self.REL_FATHER and gender == Person.MALE:
            # by male line
            return " по чоловічій лінії"
        elif by_line == self.REL_MOTHER and gender == Person.FEMALE:
            # by male line
            return " по жіночій лінії"
        elif reltocommon[0] == self.REL_FATHER:
            # From father's side
            return " по батькові"
        elif reltocommon[0] == self.REL_MOTHER:
            # From mother's side
            return " по матері"
        else:
            return ""

    def get_son(self, level, inlaw=""):
        """
        Provides information how much male descendant is related to this person
        """
        # Define if the person is natural relative or in law
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "названий "

        if level >= 0 and level < len(_son_level):
            return t_inlaw + _son_level[level]
        else:
            return t_inlaw + "пра(пра)внук у %d поколінні" % (level + 1)

    def get_daughter(self, level, inlaw=""):
        """
        Provides information how much female descendant is related to this person
        """
        # Define if the person is natural relative or in law
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "названа "

        if level >= 0 and level < len(_daughter_level):
            return t_inlaw + _daughter_level[level]
        else:
            return t_inlaw + "пра(пра)внучка у %d поколінні" % (level + 1)

    def get_child_unknown(self, level, inlaw=""):
        """
        Provides information how much descendant of unknown gender is related to this person
        """
        # Define if the person is natural relative or in law
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "названий(на) "

        if level == 1:
            return t_inlaw + "внук(а)"
        else:
            return t_inlaw + "пра(пра)внук(а) у %d поколінні" % (level + 1)

    def get_father(self, level, reltocommon, inlaw=""):
        """
        Provides information how much male ancestor (for example father)
        is related to the given person
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "названий "

        if level >= 0 and level < len(_father_level):
            # If you know exact name of relationship - use it
            if level == 1:
                # Father
                return t_inlaw + _father_level[level]
            else:
                # Grandfather, Greatgrandfather, Greatgreatgrandfather
                return (
                    t_inlaw
                    + _father_level[level]
                    + self.get_sword_distaff(level, reltocommon, Person.MALE)
                )
        else:
            # For deep generations
            return (
                t_inlaw
                + "пра(пра)дід у %d поколінні" % (level + 1)
                + self.get_sword_distaff(level, reltocommon, Person.MALE)
            )

    def get_mother(self, level, reltocommon, inlaw=""):
        """
        Provides information how much female ancestor (for example mother)
        is related to the given person
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "названа "

        if level >= 0 and level < len(_mother_level):
            # If you know exact name of relationship - use it
            if level == 1:
                # Mother
                return t_inlaw + _mother_level[level]

            else:
                # Grandmother, Greatgrandmother, Greatgreatgrandmother
                return (
                    t_inlaw
                    + _mother_level[level]
                    + self.get_sword_distaff(level, reltocommon, Person.FEMALE)
                )
        else:
            # For deep generations
            return (
                t_inlaw
                + "пра(пра)баба у %d поколінні" % (level + 1)
                + self.get_sword_distaff(level, reltocommon, Person.FEMALE)
            )

    def get_parent_unknown(self, level, reltocommon, inlaw=""):
        """
        Provides information how much an ancestor of unknown gender
        is related to the given person (unknown sex)
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "названий "

        if level == 1:
            return t_inlaw + "батько/мати"
        else:
            return (
                t_inlaw
                + "пра(пра)- дід/баба у %d поколінні" % (level + 1)
                + self.get_sword_distaff(level, reltocommon)
            )

    def get_uncle(self, level, reltocommon, inlaw=""):
        """
        Return ancle generation name
        """

        if reltocommon[0] == self.REL_FATHER:
            # from Father
            if level < len(_uncle_level):
                if level == 2:
                    return _uncle_level[level] + " (стрийко)"
                else:
                    return "двоюрідний " + _uncle_level[level] + " по батькові"
            else:
                return "двоюрідний пра(пра)дід в %d поколінні по батькові" % (level)
        elif reltocommon[0] == self.REL_MOTHER:
            # from Mother
            if level < len(_uncle_level):
                if level == 2:
                    return _uncle_level[level] + " (вуйко)"
                else:
                    return "двоюрідний " + _uncle_level[level] + " по матері"
            else:
                return "двоюрідний пра(пра)дід в %d поколінні по матері" % (level)
        else:
            if level == 2:
                return "дядько (стрийко/вуйко)"
            else:
                return "двоюрідний пра(пра)дід в %d поколінні" % (level)

    def get_aunt(self, level, reltocommon, inlaw=""):
        """
        Return aunt generation name
        """

        if reltocommon[0] == self.REL_FATHER:
            # from Father
            if level < len(_aunt_level):
                if level == 2:
                    return _aunt_level[level] + " (стрийна)"
                else:
                    return "двоюрідна " + _aunt_level[level] + " по батькові"
            else:
                return "двоюрідна пра(пра)баба в %d поколінні по батькові" % (level)
        elif reltocommon[0] == self.REL_MOTHER:
            # from Mother
            if level < len(_aunt_level):
                if level == 2:
                    return _aunt_level[level] + " (вуйна)"
                else:
                    return "двоюрідна " + _aunt_level[level] + " по матері"
            else:
                return "двоюрідна пра(пра)баба в %d поколінні по матері" % (level)
        else:
            if level == 2:
                return "тітка (стрийна/вуйна)"
            else:
                return "двоюрідна пра(пра)баба в %d поколінні" % (level)

    def get_uncle_aunt_unknown(self, level, reltocommon, inlaw=""):
        """
        Return uncle/aunt generation name when gender unknown
        """

        if reltocommon[0] == self.REL_FATHER:
            # from Father
            if level == 2:
                return "дядько/тітка (стрийко)"
            else:
                return "двоюрідний дід/баба в %d поколінні по батькові" % (level)
        elif reltocommon[0] == self.REL_MOTHER:
            # from Mother
            if level == 2:
                return "дядько/тітка (вуйко)"
            else:
                return "двоюрідний дід/баба в %d поколінні по матері" % (level)
        else:
            if level == 2:
                return "дядько/тітка"
            else:
                return "двоюрідний дід/баба в %d поколінні" % (level)

    def get_nephew(self, level, reltocommon, inlaw=""):
        """
        Return nephew generation name
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "названий "

        if level == 2:
            if reltocommon[0] == self.REL_FATHER:
                # Brother's son
                return t_inlaw + "небіж (братанець)"
            elif reltocommon[0] == self.REL_MOTHER:
                # Sister's son
                return t_inlaw + "небіж (сестринець)"
            else:
                return t_inlaw + "небіж (братанець/сестринець)"
        elif level >= 2 and level <= 6:
            if reltocommon[level - 2] == self.REL_FATHER:
                # Brother's son generation
                if reltocommon[level - 3] == self.REL_FATHER:
                    # Son of brother son
                    return self.get_son((level - 2), inlaw) + " небожа по брату"
                elif reltocommon[level - 3] == self.REL_MOTHER:
                    # Son of brother daughter
                    return self.get_son((level - 2), inlaw) + " небоги по брату"
                else:
                    return self.get_son((level - 2), inlaw) + " небожа/небоги по брату"
            elif reltocommon[level - 2] == self.REL_MOTHER:
                # Sister's son generation
                if reltocommon[level - 3] == self.REL_FATHER:
                    # Son of sister son
                    return self.get_son((level - 2), inlaw) + " небожа по сестрі"
                elif reltocommon[level - 3] == self.REL_MOTHER:
                    # Son of sister daughter
                    return self.get_son((level - 2), inlaw) + " небоги по сестрі"
                else:
                    return self.get_son((level - 2), inlaw) + " небожа/небоги по сестрі"
            else:
                return self.get_son((level - 2), inlaw) + " небожа/небоги"
        else:
            if reltocommon[level - 2] == self.REL_FATHER:
                # Brother's son generation
                return t_inlaw + "чоловічий нащадок у %d поколінні брата" % (level)
            elif reltocommon[level - 2] == self.REL_MOTHER:
                # Sister's son generation
                return t_inlaw + "чоловічий нащадок у %d поколінні сестри" % (level)
            else:
                return t_inlaw + "чоловічий нащадок у %d поколінні брата/сестри" % (
                    level
                )

    def get_niece(self, level, reltocommon, inlaw=""):
        """
        Return niece generation name
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "названий "

        if level == 2:
            if reltocommon[0] == self.REL_FATHER:
                # Brother's daughter
                return t_inlaw + "небога (братанка)"
            elif reltocommon[0] == self.REL_MOTHER:
                # Sister's daughter
                return t_inlaw + "небога (сестрениця)"
            else:
                return t_inlaw + "небога (братанка/сестрениця)"
        elif level >= 2 and level <= 6:
            if reltocommon[level - 2] == self.REL_FATHER:
                # Brother's daughter generation
                if reltocommon[level - 3] == self.REL_FATHER:
                    # daughter of brother son
                    return self.get_daughter((level - 2), inlaw) + " небожа по брату"
                elif reltocommon[level - 3] == self.REL_MOTHER:
                    # daughter of brother daughter
                    return self.get_daughter((level - 2), inlaw) + " небоги по брату"
                else:
                    return (
                        self.get_daughter((level - 2), inlaw)
                        + " небожа/небоги по брату"
                    )
            elif reltocommon[level - 2] == self.REL_MOTHER:
                # Sister's daughter generation
                if reltocommon[level - 3] == self.REL_FATHER:
                    # daughter of sister son
                    return self.get_daughter((level - 2), inlaw) + " небожа по сестрі"
                elif reltocommon[level - 3] == self.REL_MOTHER:
                    # daughter of sister daughter
                    return self.get_daughter((level - 2), inlaw) + " небоги по сестрі"
                else:
                    return (
                        self.get_daughter((level - 2), inlaw)
                        + " небожа/небоги по сестрі"
                    )
            else:
                return self.get_daughter(level, inlaw) + " небожа/небоги"
        else:
            if reltocommon[level - 2] == self.REL_FATHER:
                # Brother's daughter generation
                return t_inlaw + "жіночий нащадок у %d поколінні брата" % (level)
            elif reltocommon[level - 2] == self.REL_MOTHER:
                # Sister's daughter generation
                return t_inlaw + "жіночий нащадок у %d поколінні сестри" % (level)
            else:
                return t_inlaw + "жіночий нащадок у %d поколінні брата/сестри" % (level)

    def get_nephew_niece_unknown(self, level, reltocommon, inlaw=""):
        """
        Return nephew/niece generation name when gender unknown
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "названий "

        if reltocommon[level - 2] == self.REL_FATHER:
            # Brother's descendant
            return t_inlaw + "нащадок в %d поколінні брата" % (level)
        elif reltocommon[level - 2] == self.REL_MOTHER:
            # Sister's descendant
            return t_inlaw + "нащадок в %d поколінні сестри" % (level)
        else:
            return t_inlaw + "нащадок в %d поколінні брата aбо сестри" % (level)

    def get_level(self, level, gender):
        """
        Return level name depend of gender
        """
        if gender == Person.MALE:
            if level < len(_level_name_male):
                return _level_name_male[level]
            else:
                return "%d-юрідний" % level
        elif gender == Person.FEMALE:
            if level < len(_level_name_female):
                return _level_name_female[level]
            else:
                return "%d-юрідна" % level
        else:
            if level < len(_level_name):
                return _level_name[level]
            else:
                return "%d-юрідний(на)" % level

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Provide a string that describes the relationsip between a person, and
        another person. E.g. "grandparent" or "child".

        Ga: The number of generations between the main person and the
            common ancestor. (кількість поколінь між прямими родичами особи:
            сини, батьки, бабусі...)

        Gb: The number of generations between the other person and the
            common ancestor. (кількість поколінь між іншою особою
            і спільним предком)

        gender_a : gender of person a

        gender_b : gender of person b

        reltocommon_a : relation path to common ancestor or common
                        Family for person a.
                        Note that length = Ga
        reltocommon_b : relation path to common ancestor or common
                        Family for person b.
                        Note that length = Gb

        in_law_a : True if path to common ancestors is via the partner
                          of person a

        in_law_b : True if path to common ancestors is via the partner
                          of person b

        only_birth : True if relation between a and b is by birth only
                            False otherwise

        more see in relationship.py get_single_relationship_string()
        """

        if only_birth:
            step = ""
        else:
            step = self.STEP

        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        t_inlaw = ""

        # b is the same person as a
        if Ga == Gb == 0:
            rel_str = "та сама особа"

        elif Ga == 0:
            # b is son/descendant of a

            if gender_b == Person.MALE:
                if inlaw and Gb == 1 and not step:
                    rel_str = "зять"
                else:
                    rel_str = self.get_son(Gb, inlaw)

            elif gender_b == Person.FEMALE:
                if inlaw and Gb == 1 and not step:
                    rel_str = "невістка"
                else:
                    rel_str = self.get_daughter(Gb, inlaw)

            else:
                rel_str = self.get_child_unknown(Gb, inlaw)

        elif Gb == 0:
            # b is parent/grand parent of a

            if gender_b == Person.MALE:
                if inlaw and Gb == 1 and not step:
                    rel_str = "тесть"
                else:
                    rel_str = self.get_father(Ga, reltocommon_a, inlaw)

            elif gender_b == Person.FEMALE:
                if inlaw and Gb == 1 and not step:
                    rel_str = "теща"
                else:
                    rel_str = self.get_mother(Ga, reltocommon_a, inlaw)

            else:
                rel_str = self.get_parent_unknown(Ga, reltocommon_a, inlaw)

        elif Ga == Gb == 1:
            # Family, brother/sister
            # not used, leave it just in case
            # because should be handled
            # by get_sibling_relationship_string
            # that called in parent RelationshipCalculator
            # see Relationship.py

            if gender_b == Person.MALE:
                if inlaw and not step:
                    rel_str = "рідний брат"
                else:
                    rel_str = "напів рідний брат"

            elif gender_b == Person.FEMALE:
                if inlaw and not step:
                    rel_str = "рідна сестра"
                else:
                    rel_str = "напів рідна сестра"
            else:
                rel_str = "брат/сестра"

        elif Gb == 1 and Ga > 1:
            # b is aunt/uncle of a

            if gender_b == Person.MALE:
                rel_str = self.get_uncle(Ga, reltocommon_a, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self.get_aunt(Ga, reltocommon_a, inlaw)
            else:
                rel_str = self.get_uncle_aunt_unknown(Ga, reltocommon_a, inlaw)

        elif Ga == 1 and Gb > 1:
            # b is niece/nephew of a

            if gender_b == Person.MALE:
                rel_str = self.get_nephew(Gb, reltocommon_b, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self.get_niece(Gb, reltocommon_b, inlaw)
            else:
                rel_str = self.get_nephew_niece_unknown(Gb, reltocommon_b, inlaw)

        elif Ga > 1 and Gb > 1:
            # b is (cousin/far aunt/far  uncle/far niece/far nephew) of a

            if Ga > Gb:
                # b is far aunt/far  uncle of a
                level = Ga - Gb + 1
                if level >= 3:  # it is right or???
                    level_name = self.get_level(Gb + 1, gender_b)
                else:
                    level_name = self.get_level(Gb, gender_b)

                if gender_b == Person.MALE:
                    # b is far  uncle
                    if inlaw != "":
                        t_inlaw = "названий "
                    if level < len(_uncle_level):
                        rel_str = t_inlaw + "%s %s" % (level_name, _uncle_level[level])
                    else:
                        rel_str = t_inlaw + "%s пра(пра)дід в %d поколінні" % (
                            level_name,
                            (level),
                        )
                elif gender_b == Person.FEMALE:
                    # b is far aunt
                    if inlaw != "":
                        t_inlaw = "названа "
                    if level < len(_aunt_level):
                        rel_str = t_inlaw + "%s %s" % (level_name, _aunt_level[level])
                    else:
                        rel_str = t_inlaw + "%s пра(пра)баба в %d поколінні" % (
                            level_name,
                            (level),
                        )
                else:
                    if inlaw != "":
                        t_inlaw = "названий(на) "
                    if level == 2:
                        rel_str = t_inlaw + "%s дядько/тітка" % level_name
                    else:
                        rel_str = t_inlaw + "%s пра(пра)- дід/баба в %d поколінні" % (
                            level_name,
                            (level),
                        )

            elif Ga < Gb:
                # b is far niece/far nephew of a
                level_name = self.get_level(Ga, gender_b)
                level = Gb - Ga + 1

                if gender_b == Person.MALE:
                    # b is far nephew
                    if level == 2:
                        rel_str = "%s небіж" % level_name
                    else:
                        rel_str = "%s пра(пра)внук у %d поколінні" % (level_name, level)
                    # rel_str = "%s %s" % (level_name, self.get_nephew(level, reltocommon_b, inlaw))
                elif gender_b == Person.FEMALE:
                    # b is far niece
                    if level == 2:
                        rel_str = "%s небога" % level_name
                    else:
                        rel_str = "%s пра(пра)внучка у %d поколінні" % (
                            level_name,
                            level,
                        )
                    # rel_str = "%s %s"  % (level_name, self.get_niece(level, reltocommon_b, inlaw))
                else:
                    rel_str = "%s пра(пра)внук(а) у %d поколінні" % level
                    # rel_str = "%s %s" % (level_name, self.get_nephew_niece_unknown(level, reltocommon_b, inlaw))

            else:  # Gb == Ga
                # b is cousin of a
                level_name = self.get_level(Ga, gender_b)

                if gender_b == Person.MALE:
                    if inlaw != "":
                        t_inlaw = "названий "
                    rel_str = t_inlaw + "%s брат" % level_name
                elif gender_b == Person.FEMALE:
                    if inlaw != "":
                        t_inlaw = "названа "
                    rel_str = t_inlaw + "%s сестра" % level_name
                else:
                    if inlaw != "":
                        t_inlaw = "названий(на) "
                    rel_str = t_inlaw + "%s брат/сестра" % level_name

        else:
            # A program should never goes there, but...
            rel_str = "невизначений ступінь спорідненості"

        return rel_str

    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Generate a text with information, how far away is a group of persons
        from a main person
        """

        if Ga == Gb == 0:
            return "та сама особа"
        if 0 == Ga:
            # These are descendants
            if 1 == Gb:
                return "діти"
            if 2 == Gb:
                return "внуки"
            if 3 == Gb:
                return "правнуки"
            if 4 == Gb:
                return "праправнуки"
            return "прапрапра(n)внуки"
        if 0 == Gb:
            # These are parents/grand parents
            if 1 == Ga:
                return "батьки"
            if 2 == Ga:
                return "діди/баби"
            if 3 == Ga:
                return "прадіди/прабаби"
            if 4 == Ga:
                return "прапращури"
            return "прапрапра(n)щури"
        if 1 == Ga == Gb:
            return "батьки"
        if 1 == Gb and Ga > 1:
            return "дядьки (вуйки, стрийки) і тітки"
        if 1 < Gb and 1 == Ga:
            return "небожі по братові і сестрі"
        if 1 < Ga and 1 < Gb:
            return "дальня родина"
        return "відносини невідомі"

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        if sib_type == self.NORM_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = "рідний брат"
                elif gender_b == Person.FEMALE:
                    rel_str = "рідна сестра"
                else:
                    rel_str = "рідний(а) брат або сестра"
            else:
                if gender_b == Person.MALE:
                    rel_str = "названий брат"
                elif gender_b == Person.FEMALE:
                    rel_str = "названа сестра"
                else:
                    rel_str = "названий(а) брат або сестра"
        elif sib_type == self.UNKNOWN_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = "брат"
                elif gender_b == Person.FEMALE:
                    rel_str = "сестра"
                else:
                    rel_str = "брат або сестра"
            else:
                if gender_b == Person.MALE:
                    rel_str = "швагро"
                elif gender_b == Person.FEMALE:
                    rel_str = "братова"
                else:
                    rel_str = "швагро або братова"
        elif sib_type == self.HALF_SIB_FATHER:
            if gender_b == Person.MALE:
                rel_str = "єдинокровний(напіврідний) брат"
            elif gender_b == Person.FEMALE:
                rel_str = "єдинокровна(напіврідна) сестра"
            else:
                rel_str = "напіврідний(а) брат/сестра"
        elif sib_type == self.HALF_SIB_MOTHER:
            if gender_b == Person.MALE:
                rel_str = "єдинокровний(напіврідний) брат"
            elif gender_b == Person.FEMALE:
                rel_str = "єдинокровна(напіврідна) сестра"
            else:
                rel_str = "напіврідний(а) брат/сестра"
        elif sib_type == self.STEP_SIB:
            if gender_b == Person.MALE:
                rel_str = "зведений брат"
            elif gender_b == Person.FEMALE:
                rel_str = "зведена сестра"
            else:
                rel_str = "зведений брат або сестра"
        else:
            rel_str = "невизначена ступінь родинних відносин"
        return rel_str


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    #    python src/plugins/rel/rel_pl.py

    """
    TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test
    import signal
    import sys

    # If someone go out
    def goodby(signal, frame):
        print("No more Drink!")
        sys.exit(0)

    signal.signal(signal.SIGINT, goodby)
    # Run test
    RC = RelationshipCalculator()
    test(RC, True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Container class for managing the generic filters"""

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from xml.sax import make_parser, SAXParseException
import os
from collections import abc

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ._filterparser import FilterParser
from ..plug import BasePluginManager
from ..const import GRAMPS_LOCALE as glocale

PLUGMAN = BasePluginManager.get_instance()


# -------------------------------------------------------------------------
#
# FilterList
#
# -------------------------------------------------------------------------
class FilterList:
    """
    Container class for managing the generic filters.
    It stores, saves, and loads the filters.
    """

    def __init__(self, file):
        self.filter_namespaces = {}
        self.file = os.path.expanduser(file)
        self._cached = {}

    def get_filters_dict(self, namespace="generic"):
        """
        This runs every for every item to be matched!
        """
        if self._cached.get(namespace, None) is None:
            filters = self.get_filters(namespace)
            self._cached[namespace] = dict([(filt.name, filt) for filt in filters])
        return self._cached[namespace]

    def get_filters(self, namespace="generic"):
        """
        This runs every for every item to be matched!
        """
        if namespace in self.filter_namespaces:
            filters = self.filter_namespaces[namespace]
        else:
            filters = []
        plugins = PLUGMAN.process_plugin_data("Filters")
        if plugins:
            plugin_filters = []
            try:
                for plug in plugins:
                    if isinstance(plug, abc.Callable):
                        plug = plug(namespace)
                    if plug:
                        if isinstance(plug, (list, tuple)):
                            for subplug in plug:
                                plugin_filters.append(subplug)
                        else:
                            plugin_filters.append(plug)
            except:
                import traceback

                traceback.print_exc()
            filters += plugin_filters
        return filters

    def add(self, namespace, filt):
        """add a custom filter"""
        assert isinstance(namespace, str)

        if namespace not in self.filter_namespaces:
            self.filter_namespaces[namespace] = []
        self.filter_namespaces[namespace].append(filt)

    def load(self):
        """load a custom filter"""
        try:
            if os.path.isfile(self.file):
                parser = make_parser()
                parser.setContentHandler(FilterParser(self))
                with open(self.file, "r", encoding="utf8") as the_file:
                    parser.parse(the_file)
        except (IOError, OSError):
            print("IO/OSError in _filterlist.py")
        except SAXParseException:
            print("Parser error")

    def fix(self, line):
        """sanitize the custom filter name, if needed"""
        new_line = line.strip()
        new_line = new_line.replace("&", "&amp;")
        new_line = new_line.replace(">", "&gt;")
        new_line = new_line.replace("<", "&lt;")
        return new_line.replace('"', "&quot;")

    def save(self):
        """save the list of custom filters"""
        with open(self.file, "w", encoding="utf8") as file:
            file.write('<?xml version="1.0" encoding="utf-8"?>\n')
            file.write("<filters>\n")
            for namespace in self.filter_namespaces:
                file.write('  <object type="%s">\n' % namespace)
                filter_list = self.filter_namespaces[namespace]
                sorted_filters = sorted(
                    [(filter.get_name(), filter) for filter in filter_list],
                    key=lambda x: glocale.sort_key(x[0]),
                )
                for name, the_filter in sorted_filters:  # enable a diff
                    file.write('    <filter name="%s"' % self.fix(name))
                    file.write(' function="%s"' % the_filter.get_logical_op())
                    if the_filter.invert:
                        file.write(' invert="1"')
                    comment = the_filter.get_comment()
                    if comment:
                        file.write(' comment="%s"' % self.fix(comment))
                    file.write(">\n")
                    for rule in the_filter.get_rules():
                        file.write(
                            '      <rule class="%s" use_regex="%s" use_case="%s">'
                            "\n"
                            % (rule.__class__.__name__, rule.use_regex, rule.use_case)
                        )
                        for value in list(rule.values()):
                            file.write(
                                '        <arg value="%s"/>' "\n" % self.fix(value)
                            )
                        file.write("      </rule>\n")
                    file.write("    </filter>\n")
                file.write("  </object>\n")
            file.write("</filters>\n")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2011,2014  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
EditCitation class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.lib import Citation, NoteType, Source
from gramps.gen.db import DbTxn
from .editprimary import EditPrimary
from .objectentries import SourceEntry
from .displaytabs import NoteTab, GalleryTab, SrcAttrEmbedList, CitationBackRefList
from ..widgets import (
    MonitoredEntry,
    PrivacyButton,
    MonitoredMenu,
    MonitoredDate,
    MonitoredTagList,
)
from ..dialog import ErrorDialog
from ..glade import Glade
from gramps.gen.const import URL_MANUAL_SECT2

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT2
WIKI_HELP_SEC = _("New_Citation_dialog", "manual")

# -------------------------------------------------------------------------
#
# EditCitationclass
#
# -------------------------------------------------------------------------


class EditCitation(EditPrimary):
    """
    Create an EditCitation window. Associate a citation with the window.

    This class is called both to edit the Citation Primary object
    and to edit references from other objects to citations.
    It is called from ..editors.__init__ for editing the primary object
    and is called from CitationEmbedList for editing references

    @param callertitle: Text passed by calling object to add to title
    @type callertitle: str
    """

    def __init__(
        self, dbstate, uistate, track, obj, source=None, callback=None, callertitle=None
    ):
        """
        The obj parameter is mandatory. If the source parameter is not
        provided, it will be deduced from the obj Citation object.
        """
        if source:
            obj.set_reference_handle(source.get_handle())
        self.callertitle = callertitle
        EditPrimary.__init__(
            self,
            dbstate,
            uistate,
            track,
            obj,
            dbstate.db.get_citation_from_handle,
            dbstate.db.get_citation_from_gramps_id,
            callback,
        )

    def empty_object(self):
        """
        Return an empty Citation object for comparison for changes.

        It is used by the base class L{EditPrimary}.
        """
        return Citation()

    def get_menu_title(self):
        """
        Construct the menu title, which may include the name of the object that
        contains a reference to this citation.
        """
        title = self.obj.get_page()
        if title:
            if self.callertitle:
                title = _("Citation") + (
                    ": %(id)s - %(context)s"
                    % {"id": title, "context": self.callertitle}
                )
            else:
                title = _("Citation") + ": " + title
        else:
            if self.callertitle:
                title = _("New Citation") + (
                    ": %(id)s - %(context)s"
                    % {"id": title, "context": self.callertitle}
                )
            else:
                title = _("New Citation")
        return title

    def _local_init(self):
        """
        Local initialization function.

        Perform basic initialization, including setting up widgets
        and the glade interface. It is called by the base class L{EditPrimary},
        and overridden here.
        """
        self.glade = Glade()
        self.set_window(self.glade.toplevel, None, self.get_menu_title())
        self.setup_configs("interface.citation", 600, 450)

        self.share_btn = self.glade.get_object("select_source")
        self.add_del_btn = self.glade.get_object("add_del_source")

    def _connect_signals(self):
        """
        Connects any signals that need to be connected.

        Called by the init routine of the base class L{EditPrimary}.
        """
        self.define_ok_button(self.glade.get_object("ok"), self.save)
        self.define_cancel_button(self.glade.get_object("cancel"))
        self.define_help_button(
            self.glade.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).

        What this code does is to check that the object edited is not deleted
        whilst editing it.  If the object is deleted we need to close the editor
        windows and clean up.  If the database emits a rebuild signal for the
        database object type we also abort the edit.
        """

        self._add_db_signal("citation-rebuild", self._do_close)
        self._add_db_signal("citation-delete", self.check_for_close)
        self._add_db_signal("source-delete", self.source_delete)
        self._add_db_signal("source-update", self.source_update)

    def _setup_fields(self):
        """
        Get control widgets and attach them to Citation's attributes.
        """
        self.source_field = SourceEntry(
            self.dbstate,
            self.uistate,
            self.track,
            self.glade.get_object("source"),
            self.glade.get_object("source_event_box"),
            self.obj.set_reference_handle,
            self.obj.get_reference_handle,
            self.add_del_btn,
            self.share_btn,
            callback=self.source_changed,
        )

        self.date = MonitoredDate(
            self.glade.get_object("date_entry"),
            self.glade.get_object("date_stat"),
            self.obj.get_date_object(),
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.gid = MonitoredEntry(
            self.glade.get_object("gid"),
            self.obj.set_gramps_id,
            self.obj.get_gramps_id,
            self.db.readonly,
        )

        self.volume = MonitoredEntry(
            self.glade.get_object("volume"),
            self.obj.set_page,
            self.obj.get_page,
            self.db.readonly,
        )

        self.type_mon = MonitoredMenu(
            self.glade.get_object("confidence"),
            self.obj.set_confidence_level,
            self.obj.get_confidence_level,
            [
                (_("Very Low"), Citation.CONF_VERY_LOW),
                (_("Low"), Citation.CONF_LOW),
                (_("Normal"), Citation.CONF_NORMAL),
                (_("High"), Citation.CONF_HIGH),
                (_("Very High"), Citation.CONF_VERY_HIGH),
            ],
            self.db.readonly,
        )

        self.tags2 = MonitoredTagList(
            self.glade.get_object("tag_label"),
            self.glade.get_object("tag_button"),
            self.obj.set_tag_list,
            self.obj.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.ref_privacy = PrivacyButton(
            self.glade.get_object("privacy"), self.obj, self.db.readonly
        )

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and inserts them into the main
        window.
        """
        notebook = Gtk.Notebook()

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "citation_editor_notes",
            self.get_menu_title(),
            notetype=NoteType.CITATION,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self.gallery_tab = GalleryTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_media_list(),
        )
        self._add_tab(notebook, self.gallery_tab)
        self.track_ref_for_deletion("gallery_tab")

        self.attr_tab = SrcAttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_attribute_list(),
            "citation_editor_attributes",
        )
        self._add_tab(notebook, self.attr_tab)
        self.track_ref_for_deletion("attr_tab")

        self.citationref_list = CitationBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            self.db.find_backlink_handles(self.obj.handle),
            "citation_editor_references",
        )
        self._add_tab(notebook, self.citationref_list)
        self.track_ref_for_deletion("citationref_list")

        self._setup_notebook_tabs(notebook)

        notebook.show_all()
        self.glade.get_object("vbox").pack_start(notebook, True, True, 0)

    def source_changed(self):
        handle = self.obj.get_reference_handle()
        if handle:
            source = self.db.get_source_from_handle(handle)
            author = source.get_author()
        else:
            author = ""
        self.glade.get_object("author").set_text(author)

    def source_update(self, hndls):
        """Source changed outside of dialog, update text if its ours"""
        handle = self.obj.get_reference_handle()
        if handle and handle in hndls:
            source = self.db.get_source_from_handle(handle)
            s_lbl = "%s [%s]" % (source.get_title(), source.gramps_id)
            self.glade.get_object("source").set_text(s_lbl)
            author = source.get_author()
            self.glade.get_object("author").set_text(author)

    def source_delete(self, hndls):
        """Source deleted outside of dialog, remove it if its ours"""
        handle = self.obj.get_reference_handle()
        if handle and handle in hndls:
            self.obj.set_reference_handle(None)
            self.glade.get_object("source").set_markup(self.source_field.EMPTY_TEXT)
            self.glade.get_object("author").set_text("")
            self.source_field.set_button(False)

    def build_menu_names(self, source):
        """
        Provide the information needed by the base class to define the
        window management menu entries.
        """
        return (_("Edit Citation"), self.get_menu_title())

    def save(self, *obj):
        """
        Save the data.
        """
        self.ok_button.set_sensitive(False)
        if not self.obj.get_reference_handle():
            ErrorDialog(
                _("No source selected"),
                _(
                    "A source is anything (personal testimony, "
                    "video recording, photograph, newspaper column, "
                    "gravestone...) from which information can be "
                    "derived. To create a citation, first select the "
                    "required source, and then record the location of "
                    "the information referenced within the source in the "
                    "'Volume/Page' field."
                ),
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return

        (uses_dupe_id, gramps_id) = self._uses_duplicate_id()
        if uses_dupe_id:
            prim_object = self.get_from_gramps_id(gramps_id)
            name = prim_object.get_page()
            msg1 = _("Cannot save citation. ID already exists.")
            msg2 = _(
                "You have attempted to use the existing Gramps ID with "
                "value %(id)s. This value is already used by '"
                "%(prim_object)s'. Please enter a different ID or leave "
                "blank to get the next available ID value."
            ) % {"id": gramps_id, "prim_object": name}
            ErrorDialog(msg1, msg2, parent=self.window)
            self.ok_button.set_sensitive(True)
            return

        if not self.obj.handle:
            with DbTxn(_("Add Citation (%s)") % self.obj.get_page(), self.db) as trans:
                self.db.add_citation(self.obj, trans)
        else:
            if self.data_has_changed():
                with DbTxn(
                    _("Edit Citation (%s)") % self.obj.get_page(), self.db
                ) as trans:
                    if not self.obj.get_gramps_id():
                        self.obj.set_gramps_id(self.db.find_next_citation_gramps_id())
                    self.db.commit_citation(self.obj, trans)

        self._do_close()
        if self.callback:
            self.callback(self.obj.get_handle())

    def data_has_changed(self):
        """
        A date comparison can fail incorrectly because we have made the
        decision to store entered text in the date. However, there is no
        entered date when importing from a XML file, so we can get an
        incorrect fail.
        """
        if self.db.readonly:
            return False
        elif self.obj.handle:
            orig = self.get_from_handle(self.obj.handle)
            if orig:
                cmp_obj = orig
            else:
                cmp_obj = self.empty_object()
            return cmp_obj.serialize(True)[1:] != self.obj.serialize(True)[1:]
        else:
            cmp_obj = self.empty_object()
            return cmp_obj.serialize(True)[1:] != self.obj.serialize()[1:]



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2006  Donald N. Allingham
# Copyright (C) 2008       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Repository View
"""

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Repository
from gramps.gui.views.listview import ListView, TEXT, MARKUP, ICON
from gramps.gui.views.treemodels import RepositoryModel
from gramps.gui.views.bookmarks import RepoBookmarks
from gramps.gen.errors import WindowActiveError
from gramps.gen.config import config
from gramps.gui.editors import EditRepository
from gramps.gui.ddtargets import DdTargets
from gramps.gui.dialog import ErrorDialog
from gramps.gui.filters.sidebar import RepoSidebarFilter
from gramps.gui.merge import MergeRepository
from gramps.gen.plug import CATEGORY_QR_REPOSITORY

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# RepositoryView
#
# -------------------------------------------------------------------------
class RepositoryView(ListView):
    """repository listview class"""

    COL_NAME = 0
    COL_ID = 1
    COL_TYPE = 2
    COL_URL = 3
    COL_STREET = 4
    COL_LOCALITY = 5
    COL_CITY = 6
    COL_STATE = 7
    COL_COUNTRY = 8
    COL_ZIP = 9
    COL_EMAIL = 10
    COL_SURL = 11
    COL_PRIV = 12
    COL_TAGS = 13
    COL_CHAN = 14

    # column definitions
    COLUMNS = [
        (_("Name"), TEXT, None),
        (_("ID"), TEXT, None),
        (_("Type"), TEXT, None),
        (_("Home URL"), TEXT, None),
        (_("Street"), TEXT, None),
        (_("Locality"), TEXT, None),
        (_("City"), TEXT, None),
        (_("State/County"), TEXT, None),
        (_("Country"), TEXT, None),
        (_("ZIP/Postal Code"), TEXT, None),
        (_("Email"), TEXT, None),
        (_("Search URL"), TEXT, None),
        (_("Private"), ICON, "gramps-lock"),
        (_("Tags"), TEXT, None),
        (_("Last Changed"), TEXT, None),
    ]
    # default setting with visible columns, order of the col, and their size
    CONFIGSETTINGS = (
        (
            "columns.visible",
            [
                COL_NAME,
                COL_ID,
                COL_TYPE,
                COL_URL,
                COL_STREET,
            ],
        ),
        (
            "columns.rank",
            [
                COL_NAME,
                COL_ID,
                COL_TYPE,
                COL_URL,
                COL_STREET,
                COL_LOCALITY,
                COL_CITY,
                COL_STATE,
                COL_COUNTRY,
                COL_ZIP,
                COL_EMAIL,
                COL_SURL,
                COL_PRIV,
                COL_TAGS,
                COL_CHAN,
            ],
        ),
        (
            "columns.size",
            [200, 75, 100, 250, 100, 100, 100, 100, 100, 100, 100, 100, 40, 100, 100],
        ),
    )
    ADD_MSG = _("Add a new repository")
    EDIT_MSG = _("Edit the selected repository")
    DEL_MSG = _("Delete the selected repository")
    MERGE_MSG = _("Merge the selected repositories")
    FILTER_TYPE = "Repository"
    QR_CATEGORY = CATEGORY_QR_REPOSITORY

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        signal_map = {
            "repository-add": self.row_add,
            "repository-update": self.row_update,
            "repository-delete": self.row_delete,
            "repository-rebuild": self.object_build,
        }

        ListView.__init__(
            self,
            _("Repositories"),
            pdata,
            dbstate,
            uistate,
            RepositoryModel,
            signal_map,
            RepoBookmarks,
            nav_group,
            multiple=True,
            filter_class=RepoSidebarFilter,
        )

        self.additional_uis.append(self.additional_ui)

    def navigation_type(self):
        return "Repository"

    def drag_info(self):
        return DdTargets.REPO_LINK

    def get_stock(self):
        return "gramps-repository"

    additional_ui = [  # Defines the UI string for UIManager
        """
      <placeholder id="LocalExport">
        <item>
          <attribute name="action">win.ExportTab</attribute>
          <attribute name="label" translatable="yes">Export View...</attribute>
        </item>
      </placeholder>
""",
        """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
        % _("Organize Bookmarks"),
        """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
""",
        """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
"""
        % _("_Edit...", "action"),  # to use sgettext()
        """
        <placeholder id='otheredit'>
        <item>
          <attribute name="action">win.FilterEdit</attribute>
          <attribute name="label" translatable="yes">"""
        """Repository Filter Editor</attribute>
        </item>
        </placeholder>
""",  # Following are the Toolbar items
        """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
        """
    <placeholder id='BarCommonEdit'>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-add</property>
        <property name="action-name">win.Add</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Add...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gtk-edit</property>
        <property name="action-name">win.Edit</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">Edit...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-remove</property>
        <property name="action-name">win.Remove</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Delete</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-merge</property>
        <property name="action-name">win.Merge</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Merge...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
"""
        % (ADD_MSG, EDIT_MSG, DEL_MSG, MERGE_MSG),
        """
    <menu id="Popup">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">Forward</attribute>
        </item>
      </section>
      <section id="PopUpTree">
      </section>
      <section>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
      <section>
        <placeholder id='QuickReport'>
        </placeholder>
        <placeholder id='WebConnect'>
        </placeholder>
      </section>
    </menu>
    """
        % _("_Edit...", "action"),  # to use sgettext()
    ]

    def add(self, *obj):
        EditRepository(self.dbstate, self.uistate, [], Repository())

    def remove(self, *obj):
        """
        Method called when deleting repo(s) from the repo view.
        """
        handles = self.selected_handles()
        ht_list = [("Repository", hndl) for hndl in handles]
        self.remove_selected_objects(ht_list)

    def edit(self, *obj):
        for handle in self.selected_handles():
            repos = self.dbstate.db.get_repository_from_handle(handle)
            try:
                EditRepository(self.dbstate, self.uistate, [], repos)
            except WindowActiveError:
                pass

    def merge(self, *obj):
        """
        Merge the selected repositories.
        """
        mlist = self.selected_handles()

        if len(mlist) != 2:
            msg = _("Cannot merge repositories.")
            msg2 = _(
                "Exactly two repositories must be selected to perform a "
                "merge. A second repository can be selected by holding "
                "down the control key while clicking on the desired "
                "repository."
            )
            ErrorDialog(msg, msg2, parent=self.uistate.window)
        else:
            MergeRepository(self.dbstate, self.uistate, [], mlist[0], mlist[1])

    def get_handle_from_gramps_id(self, gid):
        obj = self.dbstate.db.get_repository_from_gramps_id(gid)
        if obj:
            return obj.get_handle()
        else:
            return None

    def tag_updated(self, handle_list):
        """
        Update tagged rows when a tag color changes.
        """
        all_links = set([])
        for tag_handle in handle_list:
            links = set(
                [
                    link[1]
                    for link in self.dbstate.db.find_backlink_handles(
                        tag_handle, include_classes="Repository"
                    )
                ]
            )
            all_links = all_links.union(links)
        self.row_update(list(all_links))

    def add_tag(self, transaction, repo_handle, tag_handle):
        """
        Add the given tag to the given repository.
        """
        repo = self.dbstate.db.get_repository_from_handle(repo_handle)
        repo.add_tag(tag_handle)
        self.dbstate.db.commit_repository(repo, transaction)

    def remove_tag(self, transaction, repo_handle, tag_handle):
        """
        Remove the given tag from the given repository.
        """
        repo = self.dbstate.db.get_repository_from_handle(repo_handle)
        repo.remove_tag(tag_handle)
        self.dbstate.db.commit_repository(repo, transaction)

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (
            ("Repository Filter",),
            ("Repository Details", "Repository Notes", "Repository Backlinks"),
        )

    def get_config_name(self):
        return __name__



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Utilities to create unique identifiers
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import sys
import random
import time
import uuid

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_UUID

# -------------------------------------------------------------------------
#
# create_id
#
# -------------------------------------------------------------------------
_rand = random.Random(time.time())
_det_id = False


def create_id():
    global _rand
    if _det_id:
        _rand = _rand + 1
        return "%08x%08x" % (_rand, _rand)
    else:
        return "%08x%08x" % (int(time.time() * 10000), _rand.randint(0, sys.maxsize))


def create_uid(self, handle=None):
    if handle:
        uid = uuid.uuid5(GRAMPS_UUID, handle)
    else:
        uid = uuid.uuid4()
    return uid.hex.upper()


def set_det_id(self, val=True):
    global _rand, _det_id
    _det_id = val
    if _det_id is True:
        _rand = 0
    else:
        _rand = random.Random(time.time())



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015-2016 Gramps Development Team
# Copyright (C) 2016      Nick Hall
# Copyright (C) 2024      Doug Blank
# Copyright (C) 2024,2025 Steve Youngs <steve@youngs.cc>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Gramps generic database handler
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
from __future__ import annotations
import bisect
import logging
import os
import pickle
import random
import re
import time
from pathlib import Path
from typing import Any, Generator, Type

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from ..errors import HandleError
from ..lib import (
    Citation,
    Event,
    Family,
    Media,
    NameOriginType,
    Note,
    Person,
    Place,
    Repository,
    Source,
    Tag,
)
from ..lib.serialize import BlobSerializer, JSONSerializer
from ..lib.genderstats import GenderStats
from ..lib.researcher import Researcher
from ..types import (
    AnyHandle,
    PersonHandle,
    EventHandle,
    FamilyHandle,
    PlaceHandle,
    PrimaryObject,
    SourceHandle,
    RepositoryHandle,
    CitationHandle,
    MediaHandle,
    NoteHandle,
    TagHandle,
    TableObjectType,
    PersonGrampsID,
    EventGrampsID,
    FamilyGrampsID,
    PlaceGrampsID,
    SourceGrampsID,
    RepositoryGrampsID,
    CitationGrampsID,
    MediaGrampsID,
    NoteGrampsID,
)
from ..updatecallback import UpdateCallback
from ..utils.callback import Callback
from ..utils.id import create_id
from . import (
    CITATION_KEY,
    DBLOGNAME,
    DBMODE_R,
    DBMODE_W,
    DBUNDOFN,
    EVENT_KEY,
    FAMILY_KEY,
    KEY_TO_NAME_MAP,
    MEDIA_KEY,
    NOTE_KEY,
    PERSON_KEY,
    PLACE_KEY,
    REFERENCE_KEY,
    REPOSITORY_KEY,
    SOURCE_KEY,
    TAG_KEY,
    TXNADD,
    TXNDEL,
    TXNUPD,
    DbReadBase,
    DbUndo,
    DbWriteBase,
)
from .bookmarks import DbBookmarks
from .exceptions import DbUpgradeRequiredError, DbVersionError
from .utils import clear_lock_file, write_lock_file
from typing import Union

_ = glocale.translation.gettext

LOG = logging.getLogger(DBLOGNAME)

SIGBASE = (
    "person",
    "family",
    "source",
    "event",
    "media",
    "place",
    "repository",
    "reference",
    "note",
    "tag",
    "citation",
)


# ------------------------------------------------------------------------
#
# DbGenericUndo class
#
# ------------------------------------------------------------------------
class DbGenericUndo(DbUndo):
    """
    Generic undo/redo handler
    """

    def __init__(self, grampsdb, path):
        super().__init__(grampsdb)
        self.undodb = []

    def open(self, value=None):
        """
        Open the backing storage.  Needs to be overridden in the derived
        class.
        """

    def close(self):
        """
        Close the backing storage.  Needs to be overridden in the derived
        class.
        """

    def append(self, value):
        """
        Add a new entry on the end.  Needs to be overridden in the derived
        class.
        """
        self.undodb.append(value)

    def __getitem__(self, index):
        """
        Returns an entry by index number.  Needs to be overridden in the
        derived class.
        """
        return self.undodb[index]

    def __setitem__(self, index, value):
        """
        Set an entry to a value.  Needs to be overridden in the derived class.
        """
        self.undodb[index] = value

    def __len__(self):
        """
        Returns the number of entries.  Needs to be overridden in the derived
        class.
        """
        return len(self.undodb)

    def _redo(self, update_history):
        """
        Access the last undone transaction, and revert the data to the state
        before the transaction was undone.
        """
        txn = self.redoq.pop()
        self.undoq.append(txn)
        transaction = txn
        db = self.db
        subitems = transaction.get_recnos()
        # sigs[obj_type][trans_type]
        sigs = [[[] for trans_type in range(3)] for key in range(11)]

        # Process all records in the transaction
        try:
            self.db._txn_begin()
            for record_id in subitems:
                (key, trans_type, handle, __, new_data) = pickle.loads(
                    self.undodb[record_id]
                )

                if key == REFERENCE_KEY:
                    self.db.undo_reference(new_data, handle)
                else:
                    self.db.undo_data(new_data, handle, key)
                    sigs[key][trans_type].append(handle)
            # now emit the signals
            self.undo_sigs(sigs, False)

            self.db._txn_commit()
        except:
            self.db._txn_abort()
            raise

        # Notify listeners
        if db.undo_callback:
            db.undo_callback(_("_Undo %s") % transaction.get_description())

        if db.redo_callback:
            if self.redo_count > 1:
                new_transaction = self.redoq[-2]
                db.redo_callback(_("_Redo %s") % new_transaction.get_description())
            else:
                db.redo_callback(None)

        if update_history and db.undo_history_callback:
            db.undo_history_callback()
        return True

    def _undo(self, update_history):
        """
        Access the last committed transaction, and revert the data to the
        state before the transaction was committed.
        """
        txn = self.undoq.pop()
        self.redoq.append(txn)
        transaction = txn
        db = self.db
        subitems = transaction.get_recnos(reverse=True)
        # sigs[obj_type][trans_type]
        sigs = [[[] for trans_type in range(3)] for key in range(11)]

        # Process all records in the transaction
        try:
            self.db._txn_begin()
            for record_id in subitems:
                (key, trans_type, handle, old_data, x) = pickle.loads(
                    self.undodb[record_id]
                )

                if key == REFERENCE_KEY:
                    self.db.undo_reference(old_data, handle)
                else:
                    self.db.undo_data(old_data, handle, key)
                    sigs[key][trans_type].append(handle)
            # now emit the signals
            self.undo_sigs(sigs, True)

            self.db._txn_commit()
        except:
            self.db._txn_abort()
            raise

        # Notify listeners
        if db.undo_callback:
            if self.undo_count > 0:
                db.undo_callback(_("_Undo %s") % self.undoq[-1].get_description())
            else:
                db.undo_callback(None)

        if db.redo_callback:
            db.redo_callback(_("_Redo %s") % transaction.get_description())

        if update_history and db.undo_history_callback:
            db.undo_history_callback()
        return True

    def undo_sigs(self, sigs, undo):
        """
        Helper method to undo/redo the signals for changes made
        We want to do deletes and adds first
        Note that if 'undo' we swap emits
        """
        for trans_type in [TXNDEL, TXNADD, TXNUPD]:
            for obj_type in range(11):
                handles = sigs[obj_type][trans_type]
                if handles:
                    if (
                        not undo
                        and trans_type == TXNDEL
                        or undo
                        and trans_type == TXNADD
                    ):
                        typ = "-delete"
                    else:
                        # don't update a handle if its been deleted, and note
                        # that 'deleted' handles are in the 'add' list if we
                        # are undoing
                        handles = [
                            handle
                            for handle in handles
                            if handle not in sigs[obj_type][TXNADD if undo else TXNDEL]
                        ]
                        if ((not undo) and trans_type == TXNADD) or (
                            undo and trans_type == TXNDEL
                        ):
                            typ = "-add"
                        else:  # TXNUPD
                            typ = "-update"
                    if handles:
                        self.db.emit(KEY_TO_NAME_MAP[obj_type] + typ, (handles,))


# ------------------------------------------------------------------------
#
# Cursor class
#
# ------------------------------------------------------------------------
class Cursor:
    """
    Simple iteration handler
    """

    def __init__(self, iterator):
        self.iterator = iterator
        self._iter = self.__iter__()

    def __enter__(self):
        return self

    def __iter__(self):
        for handle, data in self.iterator():
            yield (handle, data)

    def __next__(self):
        try:
            return self._iter.__next__()
        except StopIteration:
            return None

    def __exit__(self, *args, **kwargs):
        pass

    def iter(self):
        for handle, data in self.iterator():
            yield (handle, data)

    def first(self):
        self._iter = self.__iter__()
        try:
            return next(self._iter)
        except:
            return

    def next(self):
        try:
            return next(self._iter)
        except:
            return

    def close(self):
        pass


# ------------------------------------------------------------------------
#
# DbGeneric class
#
# ------------------------------------------------------------------------


class DbGeneric(DbWriteBase, DbReadBase, UpdateCallback, Callback):
    """
    A Gramps Database Backend. This replicates the grampsdb functions.
    """

    __signals__: dict[str, Any] = dict(
        (obj + "-" + op, signal)
        for obj in [
            "person",
            "family",
            "event",
            "place",
            "repository",
            "source",
            "citation",
            "media",
            "note",
            "tag",
        ]
        for op, signal in zip(
            ["add", "update", "delete", "rebuild"], [(list,), (list,), (list,), None]
        )
    )

    # 2. Signals for long operations
    __signals__.update(
        ("long-op-" + op, signal)
        for op, signal in zip(["start", "heartbeat", "end"], [(object,), None, None])
    )

    # 3. Special signal for change in home person
    __signals__["home-person-changed"] = None

    # 4. Signal for change in person group name, parameters are
    __signals__["person-groupname-rebuild"] = (str, str)

    __callback_map = {}

    VERSION = (21, 0, 0)

    def __init__(self, directory=None):
        DbReadBase.__init__(self)
        DbWriteBase.__init__(self)
        Callback.__init__(self)
        self.__tables = {
            "Person": {
                "handle_func": self.get_person_from_handle,
                "gramps_id_func": self.get_person_from_gramps_id,
                "class_func": Person,
                "cursor_func": self.get_person_cursor,
                "handles_func": self.get_person_handles,
                "add_func": self.add_person,
                "commit_func": self.commit_person,
                "iter_func": self.iter_people,
                "ids_func": self.get_person_gramps_ids,
                "has_handle_func": self.has_person_handle,
                "has_gramps_id_func": self.has_person_gramps_id,
                "count_func": self.get_number_of_people,
                "raw_func": self.get_raw_person_data,
                "raw_id_func": self._get_raw_person_from_id_data,
                "del_func": self.remove_person,
            },
            "Family": {
                "handle_func": self.get_family_from_handle,
                "gramps_id_func": self.get_family_from_gramps_id,
                "class_func": Family,
                "cursor_func": self.get_family_cursor,
                "handles_func": self.get_family_handles,
                "add_func": self.add_family,
                "commit_func": self.commit_family,
                "iter_func": self.iter_families,
                "ids_func": self.get_family_gramps_ids,
                "has_handle_func": self.has_family_handle,
                "has_gramps_id_func": self.has_family_gramps_id,
                "count_func": self.get_number_of_families,
                "raw_func": self.get_raw_family_data,
                "raw_id_func": self._get_raw_family_from_id_data,
                "del_func": self.remove_family,
            },
            "Source": {
                "handle_func": self.get_source_from_handle,
                "gramps_id_func": self.get_source_from_gramps_id,
                "class_func": Source,
                "cursor_func": self.get_source_cursor,
                "handles_func": self.get_source_handles,
                "add_func": self.add_source,
                "commit_func": self.commit_source,
                "iter_func": self.iter_sources,
                "ids_func": self.get_source_gramps_ids,
                "has_handle_func": self.has_source_handle,
                "has_gramps_id_func": self.has_source_gramps_id,
                "count_func": self.get_number_of_sources,
                "raw_func": self.get_raw_source_data,
                "raw_id_func": self._get_raw_source_from_id_data,
                "del_func": self.remove_source,
            },
            "Citation": {
                "handle_func": self.get_citation_from_handle,
                "gramps_id_func": self.get_citation_from_gramps_id,
                "class_func": Citation,
                "cursor_func": self.get_citation_cursor,
                "handles_func": self.get_citation_handles,
                "add_func": self.add_citation,
                "commit_func": self.commit_citation,
                "iter_func": self.iter_citations,
                "ids_func": self.get_citation_gramps_ids,
                "has_handle_func": self.has_citation_handle,
                "has_gramps_id_func": self.has_citation_gramps_id,
                "count_func": self.get_number_of_citations,
                "raw_func": self.get_raw_citation_data,
                "raw_id_func": self._get_raw_citation_from_id_data,
                "del_func": self.remove_citation,
            },
            "Event": {
                "handle_func": self.get_event_from_handle,
                "gramps_id_func": self.get_event_from_gramps_id,
                "class_func": Event,
                "cursor_func": self.get_event_cursor,
                "handles_func": self.get_event_handles,
                "add_func": self.add_event,
                "commit_func": self.commit_event,
                "iter_func": self.iter_events,
                "ids_func": self.get_event_gramps_ids,
                "has_handle_func": self.has_event_handle,
                "has_gramps_id_func": self.has_event_gramps_id,
                "count_func": self.get_number_of_events,
                "raw_func": self.get_raw_event_data,
                "raw_id_func": self._get_raw_event_from_id_data,
                "del_func": self.remove_event,
            },
            "Media": {
                "handle_func": self.get_media_from_handle,
                "gramps_id_func": self.get_media_from_gramps_id,
                "class_func": Media,
                "cursor_func": self.get_media_cursor,
                "handles_func": self.get_media_handles,
                "add_func": self.add_media,
                "commit_func": self.commit_media,
                "iter_func": self.iter_media,
                "ids_func": self.get_media_gramps_ids,
                "has_handle_func": self.has_media_handle,
                "has_gramps_id_func": self.has_media_gramps_id,
                "count_func": self.get_number_of_media,
                "raw_func": self.get_raw_media_data,
                "raw_id_func": self._get_raw_media_from_id_data,
                "del_func": self.remove_media,
            },
            "Place": {
                "handle_func": self.get_place_from_handle,
                "gramps_id_func": self.get_place_from_gramps_id,
                "class_func": Place,
                "cursor_func": self.get_place_cursor,
                "handles_func": self.get_place_handles,
                "add_func": self.add_place,
                "commit_func": self.commit_place,
                "iter_func": self.iter_places,
                "ids_func": self.get_place_gramps_ids,
                "has_handle_func": self.has_place_handle,
                "has_gramps_id_func": self.has_place_gramps_id,
                "count_func": self.get_number_of_places,
                "raw_func": self.get_raw_place_data,
                "raw_id_func": self._get_raw_place_from_id_data,
                "del_func": self.remove_place,
            },
            "Repository": {
                "handle_func": self.get_repository_from_handle,
                "gramps_id_func": self.get_repository_from_gramps_id,
                "class_func": Repository,
                "cursor_func": self.get_repository_cursor,
                "handles_func": self.get_repository_handles,
                "add_func": self.add_repository,
                "commit_func": self.commit_repository,
                "iter_func": self.iter_repositories,
                "ids_func": self.get_repository_gramps_ids,
                "has_handle_func": self.has_repository_handle,
                "has_gramps_id_func": self.has_repository_gramps_id,
                "count_func": self.get_number_of_repositories,
                "raw_func": self.get_raw_repository_data,
                "raw_id_func": self._get_raw_repository_from_id_data,
                "del_func": self.remove_repository,
            },
            "Note": {
                "handle_func": self.get_note_from_handle,
                "gramps_id_func": self.get_note_from_gramps_id,
                "class_func": Note,
                "cursor_func": self.get_note_cursor,
                "handles_func": self.get_note_handles,
                "add_func": self.add_note,
                "commit_func": self.commit_note,
                "iter_func": self.iter_notes,
                "ids_func": self.get_note_gramps_ids,
                "has_handle_func": self.has_note_handle,
                "has_gramps_id_func": self.has_note_gramps_id,
                "count_func": self.get_number_of_notes,
                "raw_func": self.get_raw_note_data,
                "raw_id_func": self._get_raw_note_from_id_data,
                "del_func": self.remove_note,
            },
            "Tag": {
                "handle_func": self.get_tag_from_handle,
                "gramps_id_func": None,
                "class_func": Tag,
                "cursor_func": self.get_tag_cursor,
                "handles_func": self.get_tag_handles,
                "add_func": self.add_tag,
                "commit_func": self.commit_tag,
                "has_handle_func": self.has_tag_handle,
                "iter_func": self.iter_tags,
                "count_func": self.get_number_of_tags,
                "raw_func": self.get_raw_tag_data,
                "del_func": self.remove_tag,
            },
        }
        self._directory = None
        self.path = None
        self.full_name = None
        self.brief_name = None
        self.readonly = False
        self.db_is_open = False
        self.name_formats = []
        # Bookmarks:
        self.bookmarks = DbBookmarks()
        self.family_bookmarks = DbBookmarks()
        self.event_bookmarks = DbBookmarks()
        self.place_bookmarks = DbBookmarks()
        self.citation_bookmarks = DbBookmarks()
        self.source_bookmarks = DbBookmarks()
        self.repo_bookmarks = DbBookmarks()
        self.media_bookmarks = DbBookmarks()
        self.note_bookmarks = DbBookmarks()
        self.set_person_id_prefix("I%04d")
        self.set_media_id_prefix("O%04d")
        self.set_family_id_prefix("F%04d")
        self.set_citation_id_prefix("C%04d")
        self.set_source_id_prefix("S%04d")
        self.set_place_id_prefix("P%04d")
        self.set_event_id_prefix("E%04d")
        self.set_repository_id_prefix("R%04d")
        self.set_note_id_prefix("N%04d")
        # Custom type values
        self.event_names = set()
        self.family_attributes = set()
        self.individual_attributes = set()
        self.source_attributes = set()
        self.marker_names = set()
        self.child_ref_types = set()
        self.family_rel_types = set()
        self.event_role_names = set()
        self.name_types = set()
        self.origin_types = set()
        self.repository_types = set()
        self.note_types = set()
        self.source_media_types = set()
        self.url_types = set()
        self.media_attributes = set()
        self.event_attributes = set()
        self.place_types = set()
        # ----------------------------------
        self.undodb = None
        self.cmap_index = 0
        self.smap_index = 0
        self.emap_index = 0
        self.pmap_index = 0
        self.fmap_index = 0
        self.lmap_index = 0
        self.omap_index = 0
        self.rmap_index = 0
        self.nmap_index = 0
        self.undo_callback = None
        self.redo_callback = None
        self.undo_history_callback = None
        self.modified = 0
        self.transaction = None
        self.abort_possible = True
        self._bm_changes = 0
        self.has_changed = 0  # Also gives commits since startup
        self.surname_list = []
        self.genderStats = GenderStats()  # can pass in loaded stats as dict
        self.owner = Researcher()
        if directory:
            self.load(directory)

    def _initialize(self, directory, username, password):
        """
        Initialize database backend.
        """
        raise NotImplementedError

    def upgrade_table_for_json_data(self, table_name):
        """
        Overload this method to upgrade the table to store data in JSON format
        """
        raise NotImplementedError

    def use_json_data(self):
        """
        Overload this method to check if the database stores objects in JSON format
        """
        raise NotImplementedError

    def __check_readonly(self, name):
        """
        Return True if we don't have read/write access to the database,
        otherwise return False (that is, we DO have read/write access)
        """
        # In-memory databases always allow write access.
        if name == ":memory:":
            return False

        # See if we write to the target directory at all?
        if not os.access(name, os.W_OK):
            return True

        # See if we lack write access to the database file
        path = os.path.join(name, "sqlite.db")
        if os.path.isfile(path) and not os.access(path, os.W_OK):
            return True

        # All tests passed.  Inform caller that we are NOT read only
        return False

    def load(
        self,
        directory,
        callback=None,
        mode=DBMODE_W,
        force_schema_upgrade=False,
        force_bsddb_upgrade=False,
        force_bsddb_downgrade=False,
        force_python_upgrade=False,
        update=True,
        username=None,
        password=None,
        json_data=True,
    ):
        """
        If update is False: then don't update any files
        """
        if self.__check_readonly(directory):
            mode = DBMODE_R

        self.readonly = mode == DBMODE_R

        if not self.readonly and directory != ":memory:":
            write_lock_file(directory)

        # run backend-specific code:
        self._initialize(directory, username, password)

        need_to_set_metadata = False
        if not self._schema_exists():
            self._create_schema(json_data)
            need_to_set_metadata = True

        if self.use_json_data():
            self.set_serializer("json")
        else:
            self.set_serializer("blob")

        if need_to_set_metadata:
            self._set_metadata("version", str(self.VERSION[0]))

        # Load metadata
        self.name_formats = self._get_metadata("name_formats")
        self.owner = self._get_metadata("researcher", default=Researcher())

        # Load bookmarks
        self.bookmarks.load(self._get_metadata("bookmarks"))
        self.family_bookmarks.load(self._get_metadata("family_bookmarks"))
        self.event_bookmarks.load(self._get_metadata("event_bookmarks"))
        self.source_bookmarks.load(self._get_metadata("source_bookmarks"))
        self.citation_bookmarks.load(self._get_metadata("citation_bookmarks"))
        self.repo_bookmarks.load(self._get_metadata("repo_bookmarks"))
        self.media_bookmarks.load(self._get_metadata("media_bookmarks"))
        self.place_bookmarks.load(self._get_metadata("place_bookmarks"))
        self.note_bookmarks.load(self._get_metadata("note_bookmarks"))

        # Custom type values
        self.event_names = self._get_metadata("event_names", set())
        self.family_attributes = self._get_metadata("fattr_names", set())
        self.individual_attributes = self._get_metadata("pattr_names", set())
        self.source_attributes = self._get_metadata("sattr_names", set())
        self.marker_names = self._get_metadata("marker_names", set())
        self.child_ref_types = self._get_metadata("child_refs", set())
        self.family_rel_types = self._get_metadata("family_rels", set())
        self.event_role_names = self._get_metadata("event_roles", set())
        self.name_types = self._get_metadata("name_types", set())
        self.origin_types = self._get_metadata("origin_types", set())
        self.repository_types = self._get_metadata("repo_types", set())
        self.note_types = self._get_metadata("note_types", set())
        self.source_media_types = self._get_metadata("sm_types", set())
        self.url_types = self._get_metadata("url_types", set())
        self.media_attributes = self._get_metadata("mattr_names", set())
        self.event_attributes = self._get_metadata("eattr_names", set())
        self.place_types = self._get_metadata("place_types", set())

        # surname list
        self.surname_list = self.get_surname_list()

        self._set_save_path(directory)

        if self._directory and self._directory != ":memory:":
            self.undolog = os.path.join(self._directory, DBUNDOFN)
        else:
            self.undolog = None
        self.undodb = self._create_undo_manager()
        self.undodb.open()

        # Other items to load
        gstats = self.get_gender_stats()
        self.genderStats = GenderStats(gstats)

        # Indexes:
        self.cmap_index = self._get_metadata("cmap_index", 0)
        self.smap_index = self._get_metadata("smap_index", 0)
        self.emap_index = self._get_metadata("emap_index", 0)
        self.pmap_index = self._get_metadata("pmap_index", 0)
        self.fmap_index = self._get_metadata("fmap_index", 0)
        self.lmap_index = self._get_metadata("lmap_index", 0)
        self.omap_index = self._get_metadata("omap_index", 0)
        self.rmap_index = self._get_metadata("rmap_index", 0)
        self.nmap_index = self._get_metadata("nmap_index", 0)

        if need_to_set_metadata:
            # for new db we always need blob metadata to allow prior Gramps versions
            # to make it to the downgrade version check
            # Note: downgrade check only works from v5.1.2 and later on sqlite
            self.set_serializer("blob")
            self.has_changed = 1  # to make sure genderstats gets saved
            self._set_all_metadata()
            if self.use_json_data():
                self.set_serializer("json")
                self._set_all_metadata()
            self.has_changed = 0  # number of commits

        self.db_is_open = True

        # Check on db version to see if we need upgrade or too new
        dbversion = int(self._get_metadata("version", default="0"))
        if dbversion > self.VERSION[0]:
            self.close()
            raise DbVersionError(dbversion, 18, self.VERSION[0])

        if not self.readonly and dbversion < self.VERSION[0]:
            LOG.debug(
                "Schema upgrade required from %s to %s", dbversion, self.VERSION[0]
            )
            if force_schema_upgrade:
                self._gramps_upgrade(dbversion, directory, callback)
            else:
                self.close(update=False)
                raise DbUpgradeRequiredError(dbversion, self.VERSION[0])

    def _create_undo_manager(self):
        """
        Create the undo manager.
        """
        return DbGenericUndo(self, self.undolog)

    def _close(self):
        """
        Close database backend.
        """
        raise NotImplementedError

    def close(self, update=True, user=None):
        """
        Close the database.
        if update is False, don't change access times, etc.
        """
        if self._directory != ":memory:":
            if update and not self.readonly:
                # This is just a dummy file to indicate last modified time of
                # the database for gramps.cli.clidbman:
                filename = os.path.join(self._directory, "meta_data.db")
                Path(filename).touch()
                self._set_all_metadata()

            self._close()

            try:
                clear_lock_file(self.get_save_path())
            except IOError:
                pass

        self.db_is_open = False
        self._directory = None

    def is_open(self):
        return self.db_is_open

    def _set_all_metadata(self):
        """
        sets all the metadata
        """
        # Save metadata
        self._set_metadata("version", str(self.VERSION[0]))
        self._set_metadata("name_formats", self.name_formats)
        self._set_metadata("researcher", self.owner)

        # Bookmarks
        self._set_metadata("bookmarks", self.bookmarks.get())
        self._set_metadata("family_bookmarks", self.family_bookmarks.get())
        self._set_metadata("event_bookmarks", self.event_bookmarks.get())
        self._set_metadata("place_bookmarks", self.place_bookmarks.get())
        self._set_metadata("repo_bookmarks", self.repo_bookmarks.get())
        self._set_metadata("source_bookmarks", self.source_bookmarks.get())
        self._set_metadata("citation_bookmarks", self.citation_bookmarks.get())
        self._set_metadata("media_bookmarks", self.media_bookmarks.get())
        self._set_metadata("note_bookmarks", self.note_bookmarks.get())

        # Custom type values, sets
        self._set_metadata("event_names", self.event_names)
        self._set_metadata("fattr_names", self.family_attributes)
        self._set_metadata("pattr_names", self.individual_attributes)
        self._set_metadata("sattr_names", self.source_attributes)
        self._set_metadata("marker_names", self.marker_names)
        self._set_metadata("child_refs", self.child_ref_types)
        self._set_metadata("family_rels", self.family_rel_types)
        self._set_metadata("event_roles", self.event_role_names)
        self._set_metadata("name_types", self.name_types)
        self._set_metadata("origin_types", self.origin_types)
        self._set_metadata("repo_types", self.repository_types)
        self._set_metadata("note_types", self.note_types)
        self._set_metadata("sm_types", self.source_media_types)
        self._set_metadata("url_types", self.url_types)
        self._set_metadata("mattr_names", self.media_attributes)
        self._set_metadata("eattr_names", self.event_attributes)
        self._set_metadata("place_types", self.place_types)

        # Save misc items:
        if self.has_changed:
            self.save_gender_stats(self.genderStats)

        # Indexes:
        self._set_metadata("cmap_index", self.cmap_index)
        self._set_metadata("smap_index", self.smap_index)
        self._set_metadata("emap_index", self.emap_index)
        self._set_metadata("pmap_index", self.pmap_index)
        self._set_metadata("fmap_index", self.fmap_index)
        self._set_metadata("lmap_index", self.lmap_index)
        self._set_metadata("omap_index", self.omap_index)
        self._set_metadata("rmap_index", self.rmap_index)
        self._set_metadata("nmap_index", self.nmap_index)

    def get_dbid(self):
        """
        We use the file directory name as the unique ID for
        this database on this computer.
        """
        return self.brief_name

    def get_dbname(self):
        """
        In DbGeneric, the database is in a text file at the path
        """
        name = None
        if self._directory:
            filepath = os.path.join(self._directory, "name.txt")
            try:
                with open(filepath, "r", encoding="utf8") as name_file:
                    name = name_file.readline().strip()
            except (OSError, IOError) as msg:
                LOG.error(str(msg))
        return name

    def version_supported(self):
        """Return True when the file has a supported version."""
        return True

    def _get_table_func(self, table=None, func=None):
        """
        Private implementation of get_table_func.
        """
        if table is None:
            return list(self.__tables.keys())
        if func is None:
            return self.__tables[table]  # dict of functions
        if func in self.__tables[table].keys():
            return self.__tables[table][func]
        return None

    def _txn_begin(self):
        """
        Lowlevel interface to the backend transaction.
        Executes a db BEGIN;
        """

    def _txn_commit(self):
        """
        Lowlevel interface to the backend transaction.
        Executes a db END;
        """

    def _txn_abort(self):
        """
        Lowlevel interface to the backend transaction.
        Executes a db ROLLBACK;
        """

    def transaction_begin(self, transaction):
        """
        Transactions are handled automatically by the db layer.
        """
        self.transaction = transaction
        return transaction

    def _get_metadata_keys(self):
        """
        Get all of the metadata setting names from the
        database.
        """
        raise NotImplementedError

    def _get_metadata(self, key, default=[]):
        """
        Get an item from the database.

        Default is an empty list, which is a mutable and
        thus a bad default (pylint will complain).

        However, it is just used as a value, and not altered, so
        its use here is ok.
        """
        raise NotImplementedError

    def _set_metadata(self, key, value, use_txn=True):
        """
        key: string
        value: item, will be serialized here

        Note: if use_txn, then begin/commit txn
        """
        raise NotImplementedError

    ################################################################
    #
    # set_*_id_prefix methods
    #
    ################################################################

    @staticmethod
    def _validated_id_prefix(val, default):
        if isinstance(val, str) and val:
            try:
                _ = val % 1
            except TypeError:  # missing conversion specifier
                prefix_var = val + "%d"
            except ValueError:  # incomplete format
                prefix_var = default + "%04d"
            else:
                prefix_var = val  # OK as given
        else:
            prefix_var = default + "%04d"  # not a string or empty string
        return prefix_var

    @staticmethod
    def __id2user_format(id_pattern):
        """
        Return a method that accepts a Gramps ID and adjusts it to the users
        format.
        """
        pattern_match = re.match(r"(.*)%[0 ](\d+)[diu]$", id_pattern)
        if pattern_match:
            str_prefix = pattern_match.group(1)

            # nr_width = int(pattern_match.group(2))
            def closure_func(gramps_id):
                if gramps_id and gramps_id.startswith(str_prefix):
                    id_number = gramps_id[len(str_prefix) :]
                    if id_number.isdigit():
                        id_value = int(id_number, 10)
                        # if len(str(id_value)) > nr_width:
                        #    # The ID to be imported is too large to fit in the
                        #    # users format. For now just create a new ID,
                        #    # because that is also what happens with IDs that
                        #    # are identical to IDs already in the database. If
                        #    # the problem of colliding import and already
                        #    # present IDs is solved the code here also needs
                        #    # some solution.
                        #    gramps_id = id_pattern % 1
                        # else:
                        gramps_id = id_pattern % id_value
                return gramps_id

        else:

            def closure_func(gramps_id):
                return gramps_id

        return closure_func

    def set_person_id_prefix(self, val):
        """
        Set the naming template for Gramps Person ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as I%d or I%04d.
        """
        self.person_prefix = self._validated_id_prefix(val, "I")
        self.id2user_format = self.__id2user_format(self.person_prefix)

    def set_citation_id_prefix(self, val):
        """
        Set the naming template for Gramps Citation ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as C%d or C%04d.
        """
        self.citation_prefix = self._validated_id_prefix(val, "C")
        self.cid2user_format = self.__id2user_format(self.citation_prefix)

    def set_source_id_prefix(self, val):
        """
        Set the naming template for Gramps Source ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as S%d or S%04d.
        """
        self.source_prefix = self._validated_id_prefix(val, "S")
        self.sid2user_format = self.__id2user_format(self.source_prefix)

    def set_media_id_prefix(self, val):
        """
        Set the naming template for Gramps Media ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as O%d or O%04d.
        """
        self.media_prefix = self._validated_id_prefix(val, "O")
        self.oid2user_format = self.__id2user_format(self.media_prefix)

    def set_place_id_prefix(self, val):
        """
        Set the naming template for Gramps Place ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as P%d or P%04d.
        """
        self.place_prefix = self._validated_id_prefix(val, "P")
        self.pid2user_format = self.__id2user_format(self.place_prefix)

    def set_family_id_prefix(self, val):
        """
        Set the naming template for Gramps Family ID values. The string is
        expected to be in the form of a simple text string, or in a format
        that contains a C/Python style format string using %d, such as F%d
        or F%04d.
        """
        self.family_prefix = self._validated_id_prefix(val, "F")
        self.fid2user_format = self.__id2user_format(self.family_prefix)

    def set_event_id_prefix(self, val):
        """
        Set the naming template for Gramps Event ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as E%d or E%04d.
        """
        self.event_prefix = self._validated_id_prefix(val, "E")
        self.eid2user_format = self.__id2user_format(self.event_prefix)

    def set_repository_id_prefix(self, val):
        """
        Set the naming template for Gramps Repository ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as R%d or R%04d.
        """
        self.repository_prefix = self._validated_id_prefix(val, "R")
        self.rid2user_format = self.__id2user_format(self.repository_prefix)

    def set_note_id_prefix(self, val):
        """
        Set the naming template for Gramps Note ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as N%d or N%04d.
        """
        self.note_prefix = self._validated_id_prefix(val, "N")
        self.nid2user_format = self.__id2user_format(self.note_prefix)

    def set_prefixes(
        self, person, media, family, source, citation, place, event, repository, note
    ):
        self.set_person_id_prefix(person)
        self.set_media_id_prefix(media)
        self.set_family_id_prefix(family)
        self.set_source_id_prefix(source)
        self.set_citation_id_prefix(citation)
        self.set_place_id_prefix(place)
        self.set_event_id_prefix(event)
        self.set_repository_id_prefix(repository)
        self.set_note_id_prefix(note)

    ################################################################
    #
    # find_next_*_gramps_id methods
    #
    ################################################################

    def _find_next_gramps_id(self, prefix, map_index, obj_key):
        """
        Helper function for find_next_<object>_gramps_id methods
        """
        index = prefix % map_index
        while self._has_gramps_id(obj_key, index):
            map_index += 1
            index = prefix % map_index
        map_index += 1
        return (map_index, index)

    def find_next_person_gramps_id(self) -> PersonGrampsID:
        """
        Return the next available GRAMPS' ID for a Person object based off the
        person ID prefix.
        """
        self.pmap_index, gid = self._find_next_gramps_id(
            self.person_prefix, self.pmap_index, PERSON_KEY
        )
        return gid

    def find_next_place_gramps_id(self) -> PlaceGrampsID:
        """
        Return the next available GRAMPS' ID for a Place object based off the
        place ID prefix.
        """
        self.lmap_index, gid = self._find_next_gramps_id(
            self.place_prefix, self.lmap_index, PLACE_KEY
        )
        return gid

    def find_next_event_gramps_id(self) -> EventGrampsID:
        """
        Return the next available GRAMPS' ID for a Event object based off the
        event ID prefix.
        """
        self.emap_index, gid = self._find_next_gramps_id(
            self.event_prefix, self.emap_index, EVENT_KEY
        )
        return gid

    def find_next_media_gramps_id(self) -> MediaGrampsID:
        """
        Return the next available GRAMPS' ID for a Media object based
        off the media object ID prefix.
        """
        self.omap_index, gid = self._find_next_gramps_id(
            self.media_prefix, self.omap_index, MEDIA_KEY
        )
        return gid

    def find_next_citation_gramps_id(self) -> CitationGrampsID:
        """
        Return the next available GRAMPS' ID for a Citation object based off the
        citation ID prefix.
        """
        self.cmap_index, gid = self._find_next_gramps_id(
            self.citation_prefix, self.cmap_index, CITATION_KEY
        )
        return gid

    def find_next_source_gramps_id(self) -> SourceGrampsID:
        """
        Return the next available GRAMPS' ID for a Source object based off the
        source ID prefix.
        """
        self.smap_index, gid = self._find_next_gramps_id(
            self.source_prefix, self.smap_index, SOURCE_KEY
        )
        return gid

    def find_next_family_gramps_id(self) -> FamilyGrampsID:
        """
        Return the next available GRAMPS' ID for a Family object based off the
        family ID prefix.
        """
        self.fmap_index, gid = self._find_next_gramps_id(
            self.family_prefix, self.fmap_index, FAMILY_KEY
        )
        return gid

    def find_next_repository_gramps_id(self) -> RepositoryGrampsID:
        """
        Return the next available GRAMPS' ID for a Respository object based
        off the repository ID prefix.
        """
        self.rmap_index, gid = self._find_next_gramps_id(
            self.repository_prefix, self.rmap_index, REPOSITORY_KEY
        )
        return gid

    def find_next_note_gramps_id(self) -> NoteGrampsID:
        """
        Return the next available GRAMPS' ID for a Note object based off the
        note ID prefix.
        """
        self.nmap_index, gid = self._find_next_gramps_id(
            self.note_prefix, self.nmap_index, NOTE_KEY
        )
        return gid

    ################################################################
    #
    # get_number_of_* methods
    #
    ################################################################

    def _get_number_of(self, obj_key):
        """
        Return the number of objects currently in the database.
        """
        raise NotImplementedError

    def get_number_of_people(self):
        """
        Return the number of people currently in the database.
        """
        return self._get_number_of(PERSON_KEY)

    def get_number_of_events(self):
        """
        Return the number of events currently in the database.
        """
        return self._get_number_of(EVENT_KEY)

    def get_number_of_places(self):
        """
        Return the number of places currently in the database.
        """
        return self._get_number_of(PLACE_KEY)

    def get_number_of_tags(self):
        """
        Return the number of tags currently in the database.
        """
        return self._get_number_of(TAG_KEY)

    def get_number_of_families(self):
        """
        Return the number of families currently in the database.
        """
        return self._get_number_of(FAMILY_KEY)

    def get_number_of_notes(self):
        """
        Return the number of notes currently in the database.
        """
        return self._get_number_of(NOTE_KEY)

    def get_number_of_citations(self):
        """
        Return the number of citations currently in the database.
        """
        return self._get_number_of(CITATION_KEY)

    def get_number_of_sources(self):
        """
        Return the number of sources currently in the database.
        """
        return self._get_number_of(SOURCE_KEY)

    def get_number_of_media(self):
        """
        Return the number of media objects currently in the database.
        """
        return self._get_number_of(MEDIA_KEY)

    def get_number_of_repositories(self):
        """
        Return the number of source repositories currently in the database.
        """
        return self._get_number_of(REPOSITORY_KEY)

    ################################################################
    #
    # get_*_gramps_ids methods
    #
    ################################################################

    def _get_gramps_ids(self, obj_key):
        """
        Return a list of Gramps IDs, one ID for each object in the
        database.
        """
        raise NotImplementedError

    def get_person_gramps_ids(self):
        """
        Return a list of Gramps IDs, one ID for each Person in the
        database.
        """
        return self._get_gramps_ids(PERSON_KEY)

    def get_family_gramps_ids(self):
        """
        Return a list of Gramps IDs, one ID for each Family in the
        database.
        """
        return self._get_gramps_ids(FAMILY_KEY)

    def get_source_gramps_ids(self):
        """
        Return a list of Gramps IDs, one ID for each Source in the
        database.
        """
        return self._get_gramps_ids(SOURCE_KEY)

    def get_citation_gramps_ids(self):
        """
        Return a list of Gramps IDs, one ID for each Citation in the
        database.
        """
        return self._get_gramps_ids(CITATION_KEY)

    def get_event_gramps_ids(self):
        """
        Return a list of Gramps IDs, one ID for each Event in the
        database.
        """
        return self._get_gramps_ids(EVENT_KEY)

    def get_media_gramps_ids(self):
        """
        Return a list of Gramps IDs, one ID for each Media in the
        database.
        """
        return self._get_gramps_ids(MEDIA_KEY)

    def get_place_gramps_ids(self):
        """
        Return a list of Gramps IDs, one ID for each Place in the
        database.
        """
        return self._get_gramps_ids(PLACE_KEY)

    def get_repository_gramps_ids(self):
        """
        Return a list of Gramps IDs, one ID for each Repository in the
        database.
        """
        return self._get_gramps_ids(REPOSITORY_KEY)

    def get_note_gramps_ids(self):
        """
        Return a list of Gramps IDs, one ID for each Note in the
        database.
        """
        return self._get_gramps_ids(NOTE_KEY)

    ################################################################
    #
    # get_*_from_handle methods
    #
    ################################################################

    def _get_from_handle(
        self, obj_key, obj_class: Type[TableObjectType], handle: AnyHandle
    ) -> TableObjectType:
        if handle is None:
            raise HandleError("Handle is None")
        if not handle:
            raise HandleError("Handle is empty")
        data = self._get_raw_data(obj_key, handle)
        if data:
            return self.serializer.data_to_object(data, obj_class)

        raise HandleError(f"Handle {handle} not found")

    def get_event_from_handle(self, handle: EventHandle) -> Event:
        return self._get_from_handle(EVENT_KEY, Event, handle)

    def get_family_from_handle(self, handle: FamilyHandle) -> Family:
        return self._get_from_handle(FAMILY_KEY, Family, handle)

    def get_repository_from_handle(self, handle: RepositoryHandle) -> Repository:
        return self._get_from_handle(REPOSITORY_KEY, Repository, handle)

    def get_person_from_handle(self, handle: PersonHandle) -> Person:
        return self._get_from_handle(PERSON_KEY, Person, handle)

    def get_place_from_handle(self, handle: PlaceHandle) -> Place:
        return self._get_from_handle(PLACE_KEY, Place, handle)

    def get_citation_from_handle(self, handle: CitationHandle) -> Citation:
        return self._get_from_handle(CITATION_KEY, Citation, handle)

    def get_source_from_handle(self, handle: SourceHandle) -> Source:
        return self._get_from_handle(SOURCE_KEY, Source, handle)

    def get_note_from_handle(self, handle: NoteHandle) -> Note:
        return self._get_from_handle(NOTE_KEY, Note, handle)

    def get_media_from_handle(self, handle: MediaHandle) -> Media:
        return self._get_from_handle(MEDIA_KEY, Media, handle)

    def get_tag_from_handle(self, handle: TagHandle) -> Tag:
        return self._get_from_handle(TAG_KEY, Tag, handle)

    ################################################################
    #
    # get_*_from_gramps_id methods
    #
    ################################################################

    def get_person_from_gramps_id(self, gramps_id: PersonGrampsID) -> Person:
        data = self._get_raw_person_from_id_data(gramps_id)
        return self.serializer.data_to_object(data, Person)

    def get_family_from_gramps_id(self, gramps_id: FamilyGrampsID) -> Family:
        data = self._get_raw_family_from_id_data(gramps_id)
        return self.serializer.data_to_object(data, Family)

    def get_citation_from_gramps_id(self, gramps_id: CitationGrampsID) -> Citation:
        data = self._get_raw_citation_from_id_data(gramps_id)
        return self.serializer.data_to_object(data, Citation)

    def get_source_from_gramps_id(self, gramps_id: SourceGrampsID) -> Source:
        data = self._get_raw_source_from_id_data(gramps_id)
        return self.serializer.data_to_object(data, Source)

    def get_event_from_gramps_id(self, gramps_id: EventGrampsID) -> Event:
        data = self._get_raw_event_from_id_data(gramps_id)
        return self.serializer.data_to_object(data, Event)

    def get_media_from_gramps_id(self, gramps_id: MediaGrampsID) -> Media:
        data = self._get_raw_media_from_id_data(gramps_id)
        return self.serializer.data_to_object(data, Media)

    def get_place_from_gramps_id(self, gramps_id: PlaceGrampsID) -> Place:
        data = self._get_raw_place_from_id_data(gramps_id)
        return self.serializer.data_to_object(data, Place)

    def get_repository_from_gramps_id(
        self, gramps_id: RepositoryGrampsID
    ) -> Repository:
        data = self._get_raw_repository_from_id_data(gramps_id)
        return self.serializer.data_to_object(data, Repository)

    def get_note_from_gramps_id(self, gramps_id: NoteGrampsID) -> Note:
        data = self._get_raw_note_from_id_data(gramps_id)
        return self.serializer.data_to_object(data, Note)

    ################################################################
    #
    # has_*_handle methods
    #
    ################################################################

    def _has_handle(self, obj_key, handle) -> bool:
        """
        Return True if the handle exists in the database.
        """
        raise NotImplementedError

    def has_person_handle(self, handle: PersonHandle) -> bool:
        return self._has_handle(PERSON_KEY, handle)

    def has_family_handle(self, handle: FamilyHandle) -> bool:
        return self._has_handle(FAMILY_KEY, handle)

    def has_source_handle(self, handle: SourceHandle) -> bool:
        return self._has_handle(SOURCE_KEY, handle)

    def has_citation_handle(self, handle: CitationHandle) -> bool:
        return self._has_handle(CITATION_KEY, handle)

    def has_event_handle(self, handle: EventHandle) -> bool:
        return self._has_handle(EVENT_KEY, handle)

    def has_media_handle(self, handle: MediaHandle) -> bool:
        return self._has_handle(MEDIA_KEY, handle)

    def has_place_handle(self, handle: PlaceHandle) -> bool:
        return self._has_handle(PLACE_KEY, handle)

    def has_repository_handle(self, handle: RepositoryHandle) -> bool:
        return self._has_handle(REPOSITORY_KEY, handle)

    def has_note_handle(self, handle: NoteHandle) -> bool:
        return self._has_handle(NOTE_KEY, handle)

    def has_tag_handle(self, handle: TagHandle) -> bool:
        return self._has_handle(TAG_KEY, handle)

    ################################################################
    #
    # has_*_gramps_id methods
    #
    ################################################################

    def _has_gramps_id(self, obj_key, gramps_id) -> bool:
        raise NotImplementedError

    def has_person_gramps_id(self, gramps_id: PersonGrampsID) -> bool:
        return self._has_gramps_id(PERSON_KEY, gramps_id)

    def has_family_gramps_id(self, gramps_id: FamilyGrampsID) -> bool:
        return self._has_gramps_id(FAMILY_KEY, gramps_id)

    def has_source_gramps_id(self, gramps_id: SourceGrampsID) -> bool:
        return self._has_gramps_id(SOURCE_KEY, gramps_id)

    def has_citation_gramps_id(self, gramps_id: CitationGrampsID) -> bool:
        return self._has_gramps_id(CITATION_KEY, gramps_id)

    def has_event_gramps_id(self, gramps_id: EventGrampsID) -> bool:
        return self._has_gramps_id(EVENT_KEY, gramps_id)

    def has_media_gramps_id(self, gramps_id: MediaGrampsID) -> bool:
        return self._has_gramps_id(MEDIA_KEY, gramps_id)

    def has_place_gramps_id(self, gramps_id: PlaceGrampsID) -> bool:
        return self._has_gramps_id(PLACE_KEY, gramps_id)

    def has_repository_gramps_id(self, gramps_id: RepositoryGrampsID) -> bool:
        return self._has_gramps_id(REPOSITORY_KEY, gramps_id)

    def has_note_gramps_id(self, gramps_id: NoteGrampsID) -> bool:
        return self._has_gramps_id(NOTE_KEY, gramps_id)

    ################################################################
    #
    # get_*_cursor methods
    #
    ################################################################

    def get_place_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_place_data)

    def get_place_tree_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_place_tree_data)

    def get_person_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_person_data)

    def get_family_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_family_data)

    def get_event_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_event_data)

    def get_note_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_note_data)

    def get_tag_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_tag_data)

    def get_repository_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_repository_data)

    def get_media_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_media_data)

    def get_citation_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_citation_data)

    def get_source_cursor(self) -> Cursor:
        return Cursor(self._iter_raw_source_data)

    ################################################################
    #
    # iter_*_handles methods
    #
    ################################################################

    def _iter_handles(self, obj_key):
        raise NotImplementedError

    def iter_person_handles(self) -> Generator[PersonHandle]:
        """
        Return an iterator over handles for Persons in the database
        """
        return self._iter_handles(PERSON_KEY)

    def iter_family_handles(self) -> Generator[FamilyHandle]:
        """
        Return an iterator over handles for Families in the database
        """
        return self._iter_handles(FAMILY_KEY)

    def iter_citation_handles(self) -> Generator[CitationHandle]:
        """
        Return an iterator over database handles, one handle for each Citation
        in the database.
        """
        return self._iter_handles(CITATION_KEY)

    def iter_event_handles(self) -> Generator[EventHandle]:
        """
        Return an iterator over handles for Events in the database
        """
        return self._iter_handles(EVENT_KEY)

    def iter_media_handles(self) -> Generator[MediaHandle]:
        """
        Return an iterator over handles for Media in the database
        """
        return self._iter_handles(MEDIA_KEY)

    def iter_note_handles(self) -> Generator[NoteHandle]:
        """
        Return an iterator over handles for Notes in the database
        """
        return self._iter_handles(NOTE_KEY)

    def iter_place_handles(self) -> Generator[PlaceHandle]:
        """
        Return an iterator over handles for Places in the database
        """
        return self._iter_handles(PLACE_KEY)

    def iter_repository_handles(self) -> Generator[RepositoryHandle]:
        """
        Return an iterator over handles for Repositories in the database
        """
        return self._iter_handles(REPOSITORY_KEY)

    def iter_source_handles(self) -> Generator[SourceHandle]:
        """
        Return an iterator over handles for Sources in the database
        """
        return self._iter_handles(SOURCE_KEY)

    def iter_tag_handles(self) -> Generator[TagHandle]:
        """
        Return an iterator over handles for Tags in the database
        """
        return self._iter_handles(TAG_KEY)

    ################################################################
    #
    # iter_* methods
    #
    ################################################################

    def _iter_objects(self, class_):
        """
        Iterate over items in a class.
        """
        cursor = self._get_table_func(class_.__name__, "cursor_func")
        for handle, data in cursor():
            yield self.serializer.data_to_object(data, class_)

    def iter_people(self) -> Generator[Person]:
        """Iterate over Person objects."""
        return self._iter_objects(Person)

    def iter_families(self) -> Generator[Family]:
        """Iterate over Family objects."""
        return self._iter_objects(Family)

    def iter_citations(self) -> Generator[Citation]:
        """Iterate over Citation objects."""
        return self._iter_objects(Citation)

    def iter_events(self) -> Generator[Event]:
        """Iterate over Event objects."""
        return self._iter_objects(Event)

    def iter_media(self) -> Generator[Media]:
        """Iterate over Media objects."""
        return self._iter_objects(Media)

    def iter_notes(self) -> Generator[Note]:
        """Iterate over Note objects."""
        return self._iter_objects(Note)

    def iter_places(self) -> Generator[Place]:
        """Iterate over Place objects."""
        return self._iter_objects(Place)

    def iter_repositories(self) -> Generator[Repository]:
        """Iterate over Repository objects."""
        return self._iter_objects(Repository)

    def iter_sources(self) -> Generator[Source]:
        """Iterate over Source objects."""
        return self._iter_objects(Source)

    def iter_tags(self) -> Generator[Tag]:
        """Iterate over Tag objects."""
        return self._iter_objects(Tag)

    ################################################################
    #
    # _iter_raw_*_data methods
    #
    ################################################################

    def _iter_raw_data(self, obj_key):
        raise NotImplementedError

    def _iter_raw_person_data(self):
        """
        Return an iterator over raw Person data.
        """
        return self._iter_raw_data(PERSON_KEY)

    def _iter_raw_family_data(self):
        """
        Return an iterator over raw Family data.
        """
        return self._iter_raw_data(FAMILY_KEY)

    def _iter_raw_event_data(self):
        """
        Return an iterator over raw Event data.
        """
        return self._iter_raw_data(EVENT_KEY)

    def _iter_raw_place_data(self):
        """
        Return an iterator over raw Place data.
        """
        return self._iter_raw_data(PLACE_KEY)

    def _iter_raw_repository_data(self):
        """
        Return an iterator over raw Repository data.
        """
        return self._iter_raw_data(REPOSITORY_KEY)

    def _iter_raw_source_data(self):
        """
        Return an iterator over raw Source data.
        """
        return self._iter_raw_data(SOURCE_KEY)

    def _iter_raw_citation_data(self):
        """
        Return an iterator over raw Citation data.
        """
        return self._iter_raw_data(CITATION_KEY)

    def _iter_raw_media_data(self):
        """
        Return an iterator over raw Media data.
        """
        return self._iter_raw_data(MEDIA_KEY)

    def _iter_raw_note_data(self):
        """
        Return an iterator over raw Note data.
        """
        return self._iter_raw_data(NOTE_KEY)

    def _iter_raw_tag_data(self):
        """
        Return an iterator over raw Tag data.
        """
        return self._iter_raw_data(TAG_KEY)

    def _iter_raw_place_tree_data(self):
        """
        Return an iterator over raw data in the place hierarchy.
        """
        raise NotImplementedError

    ################################################################
    #
    # get_raw_*_data methods
    #
    ################################################################

    def _get_raw_data(self, obj_key, handle):
        """
        Return raw (serialized) object from handle.
        """
        raise NotImplementedError

    def get_raw_person_data(self, handle):
        return self._get_raw_data(PERSON_KEY, handle)

    def get_raw_family_data(self, handle):
        return self._get_raw_data(FAMILY_KEY, handle)

    def get_raw_source_data(self, handle):
        return self._get_raw_data(SOURCE_KEY, handle)

    def get_raw_citation_data(self, handle):
        return self._get_raw_data(CITATION_KEY, handle)

    def get_raw_event_data(self, handle):
        return self._get_raw_data(EVENT_KEY, handle)

    def get_raw_media_data(self, handle):
        return self._get_raw_data(MEDIA_KEY, handle)

    def get_raw_place_data(self, handle):
        return self._get_raw_data(PLACE_KEY, handle)

    def get_raw_repository_data(self, handle):
        return self._get_raw_data(REPOSITORY_KEY, handle)

    def get_raw_note_data(self, handle):
        return self._get_raw_data(NOTE_KEY, handle)

    def get_raw_tag_data(self, handle):
        return self._get_raw_data(TAG_KEY, handle)

    ################################################################
    #
    # get_raw_*_from_id_data methods
    #
    ################################################################

    def _get_raw_from_id_data(self, obj_key, gramps_id):
        raise NotImplementedError

    def _get_raw_person_from_id_data(self, gramps_id: PersonGrampsID):
        return self._get_raw_from_id_data(PERSON_KEY, gramps_id)

    def _get_raw_family_from_id_data(self, gramps_id: FamilyGrampsID):
        return self._get_raw_from_id_data(FAMILY_KEY, gramps_id)

    def _get_raw_source_from_id_data(self, gramps_id: SourceGrampsID):
        return self._get_raw_from_id_data(SOURCE_KEY, gramps_id)

    def _get_raw_citation_from_id_data(self, gramps_id: CitationGrampsID):
        return self._get_raw_from_id_data(CITATION_KEY, gramps_id)

    def _get_raw_event_from_id_data(self, gramps_id: EventGrampsID):
        return self._get_raw_from_id_data(EVENT_KEY, gramps_id)

    def _get_raw_media_from_id_data(self, gramps_id: MediaGrampsID):
        return self._get_raw_from_id_data(MEDIA_KEY, gramps_id)

    def _get_raw_place_from_id_data(self, gramps_id: PlaceGrampsID):
        return self._get_raw_from_id_data(PLACE_KEY, gramps_id)

    def _get_raw_repository_from_id_data(self, gramps_id: RepositoryGrampsID):
        return self._get_raw_from_id_data(REPOSITORY_KEY, gramps_id)

    def _get_raw_note_from_id_data(self, gramps_id: NoteGrampsID):
        return self._get_raw_from_id_data(NOTE_KEY, gramps_id)

    ################################################################
    #
    # add_* methods
    #
    ################################################################

    def _add_base(self, obj, trans, set_gid: bool, find_func, commit_func):
        if not obj.handle:
            obj.handle = create_id()
        if (not obj.gramps_id) and set_gid:
            obj.gramps_id = find_func()
        if not obj.gramps_id:
            # give it a random value for the moment:
            obj.gramps_id = str(random.random())
        commit_func(obj, trans)
        return obj.handle

    def add_person(self, person: Person, transaction, set_gid=True):
        return self._add_base(
            person,
            transaction,
            set_gid,
            self.find_next_person_gramps_id,
            self.commit_person,
        )

    def add_family(self, family: Family, transaction, set_gid=True):
        return self._add_base(
            family,
            transaction,
            set_gid,
            self.find_next_family_gramps_id,
            self.commit_family,
        )

    def add_event(self, event: Event, transaction, set_gid=True):
        return self._add_base(
            event,
            transaction,
            set_gid,
            self.find_next_event_gramps_id,
            self.commit_event,
        )

    def add_place(self, place: Place, transaction, set_gid=True):
        return self._add_base(
            place,
            transaction,
            set_gid,
            self.find_next_place_gramps_id,
            self.commit_place,
        )

    def add_repository(self, repository: Repository, transaction, set_gid=True):
        return self._add_base(
            repository,
            transaction,
            set_gid,
            self.find_next_repository_gramps_id,
            self.commit_repository,
        )

    def add_source(self, source: Source, transaction, set_gid=True):
        return self._add_base(
            source,
            transaction,
            set_gid,
            self.find_next_source_gramps_id,
            self.commit_source,
        )

    def add_citation(self, citation: Citation, transaction, set_gid=True):
        return self._add_base(
            citation,
            transaction,
            set_gid,
            self.find_next_citation_gramps_id,
            self.commit_citation,
        )

    def add_media(self, media: Media, transaction, set_gid=True):
        return self._add_base(
            media,
            transaction,
            set_gid,
            self.find_next_media_gramps_id,
            self.commit_media,
        )

    def add_note(self, note: Note, transaction, set_gid=True):
        return self._add_base(
            note, transaction, set_gid, self.find_next_note_gramps_id, self.commit_note
        )

    def add_tag(self, tag: Tag, transaction):
        if not tag.handle:
            tag.handle = create_id()
        self.commit_tag(tag, transaction)
        return tag.handle

    ################################################################
    #
    # commit_* methods
    #
    ################################################################

    def _commit_base(self, obj, obj_key, trans, change_time):
        """
        Commit the specified object to the database, storing the changes as
        part of the transaction.
        """
        raise NotImplementedError

    def commit_person(self, person, transaction, change_time=None):
        """
        Commit the specified Person to the database, storing the changes as
        part of the transaction.
        """
        old_data = self._commit_base(person, PERSON_KEY, transaction, change_time)

        if old_data:
            old_person = self.serializer.data_to_object(old_data, Person)
            # Update gender statistics if necessary
            if old_person.gender != person.gender or (
                old_person.primary_name.first_name != person.primary_name.first_name
            ):
                self.genderStats.uncount_person(old_person)
                self.genderStats.count_person(person)
            # Update surname list if necessary
            if self._order_by_person_key(person) != self._order_by_person_key(
                old_person
            ):
                self.remove_from_surname_list(old_person)
                self.add_to_surname_list(person, transaction.batch)
        else:
            self.genderStats.count_person(person)
            self.add_to_surname_list(person, transaction.batch)

        # Other misc update tasks:
        self.individual_attributes.update(
            [
                str(attr.type)
                for attr in person.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        )

        self.event_role_names.update(
            [str(eref.role) for eref in person.event_ref_list if eref.role.is_custom()]
        )

        self.name_types.update(
            [
                str(name.type)
                for name in ([person.primary_name] + person.alternate_names)
                if name.type.is_custom()
            ]
        )
        all_surn = []  # new list we will use for storage
        all_surn += person.primary_name.get_surname_list()
        for asurname in person.alternate_names:
            all_surn += asurname.get_surname_list()
        self.origin_types.update(
            [str(surn.origintype) for surn in all_surn if surn.origintype.is_custom()]
        )
        all_surn = None
        self.url_types.update(
            [str(url.type) for url in person.urls if url.type.is_custom()]
        )
        attr_list = []
        for mref in person.media_list:
            attr_list += [
                str(attr.type)
                for attr in mref.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        self.media_attributes.update(attr_list)

    def commit_family(self, family, transaction, change_time=None):
        """
        Commit the specified Family to the database, storing the changes as
        part of the transaction.
        """
        self._commit_base(family, FAMILY_KEY, transaction, change_time)

        # Misc updates:
        self.family_attributes.update(
            [
                str(attr.type)
                for attr in family.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        )

        rel_list = []
        for ref in family.child_ref_list:
            if ref.frel.is_custom():
                rel_list.append(str(ref.frel))
            if ref.mrel.is_custom():
                rel_list.append(str(ref.mrel))
        self.child_ref_types.update(rel_list)

        self.event_role_names.update(
            [str(eref.role) for eref in family.event_ref_list if eref.role.is_custom()]
        )

        if family.type.is_custom():
            self.family_rel_types.add(str(family.type))

        attr_list = []
        for mref in family.media_list:
            attr_list += [
                str(attr.type)
                for attr in mref.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        self.media_attributes.update(attr_list)

    def commit_citation(self, citation, transaction, change_time=None):
        """
        Commit the specified Citation to the database, storing the changes as
        part of the transaction.
        """
        self._commit_base(citation, CITATION_KEY, transaction, change_time)

        # Misc updates:
        attr_list = []
        for mref in citation.media_list:
            attr_list += [
                str(attr.type)
                for attr in mref.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        self.media_attributes.update(attr_list)

        self.source_attributes.update(
            [
                str(attr.type)
                for attr in citation.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        )

    def commit_source(self, source, transaction, change_time=None):
        """
        Commit the specified Source to the database, storing the changes as
        part of the transaction.
        """
        self._commit_base(source, SOURCE_KEY, transaction, change_time)

        # Misc updates:
        self.source_media_types.update(
            [
                str(ref.media_type)
                for ref in source.reporef_list
                if ref.media_type.is_custom()
            ]
        )

        attr_list = []
        for mref in source.media_list:
            attr_list += [
                str(attr.type)
                for attr in mref.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        self.media_attributes.update(attr_list)
        self.source_attributes.update(
            [
                str(attr.type)
                for attr in source.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        )

    def commit_repository(self, repository, transaction, change_time=None):
        """
        Commit the specified Repository to the database, storing the changes
        as part of the transaction.
        """
        self._commit_base(repository, REPOSITORY_KEY, transaction, change_time)

        # Misc updates:
        if repository.type.is_custom():
            self.repository_types.add(str(repository.type))

        self.url_types.update(
            [str(url.type) for url in repository.urls if url.type.is_custom()]
        )

    def commit_note(self, note, transaction, change_time=None):
        """
        Commit the specified Note to the database, storing the changes as part
        of the transaction.
        """
        self._commit_base(note, NOTE_KEY, transaction, change_time)

        # Misc updates:
        if note.type.is_custom():
            self.note_types.add(str(note.type))

    def commit_place(self, place, transaction, change_time=None):
        """
        Commit the specified Place to the database, storing the changes as
        part of the transaction.
        """
        self._commit_base(place, PLACE_KEY, transaction, change_time)

        # Misc updates:
        if place.get_type().is_custom():
            self.place_types.add(str(place.get_type()))

        self.url_types.update(
            [str(url.type) for url in place.urls if url.type.is_custom()]
        )

        attr_list = []
        for mref in place.media_list:
            attr_list += [
                str(attr.type)
                for attr in mref.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        self.media_attributes.update(attr_list)

    def commit_event(self, event, transaction, change_time=None):
        """
        Commit the specified Event to the database, storing the changes as
        part of the transaction.
        """
        self._commit_base(event, EVENT_KEY, transaction, change_time)

        # Misc updates:
        self.event_attributes.update(
            [
                str(attr.type)
                for attr in event.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        )
        if event.type.is_custom():
            self.event_names.add(str(event.type))
        attr_list = []
        for mref in event.media_list:
            attr_list += [
                str(attr.type)
                for attr in mref.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        self.media_attributes.update(attr_list)

    def commit_tag(self, tag, transaction, change_time=None):
        """
        Commit the specified Tag to the database, storing the changes as
        part of the transaction.
        """
        self._commit_base(tag, TAG_KEY, transaction, change_time)

    def commit_media(self, media, transaction, change_time=None):
        """
        Commit the specified Media to the database, storing the changes
        as part of the transaction.
        """
        self._commit_base(media, MEDIA_KEY, transaction, change_time)

        # Misc updates:
        self.media_attributes.update(
            [
                str(attr.type)
                for attr in media.attribute_list
                if attr.type.is_custom() and str(attr.type)
            ]
        )

    def _after_commit(self, transaction):
        """
        Post-transaction commit processing
        """
        # Reset callbacks if necessary
        if transaction.batch or not len(transaction):
            return
        if self.undo_callback:
            self.undo_callback(_("_Undo %s") % transaction.get_description())
        if self.redo_callback:
            self.redo_callback(None)
        if self.undo_history_callback:
            self.undo_history_callback()

    ################################################################
    #
    # remove_* methods
    #
    ################################################################

    def _do_remove(self, handle, transaction, obj_key):
        raise NotImplementedError

    def remove_person(self, handle, transaction):
        """
        Remove the Person specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, PERSON_KEY)

    def remove_source(self, handle, transaction):
        """
        Remove the Source specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, SOURCE_KEY)

    def remove_citation(self, handle, transaction):
        """
        Remove the Citation specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, CITATION_KEY)

    def remove_event(self, handle, transaction):
        """
        Remove the Event specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, EVENT_KEY)

    def remove_media(self, handle, transaction):
        """
        Remove the MediaPerson specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, MEDIA_KEY)

    def remove_place(self, handle, transaction):
        """
        Remove the Place specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, PLACE_KEY)

    def remove_family(self, handle, transaction):
        """
        Remove the Family specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, FAMILY_KEY)

    def remove_repository(self, handle, transaction):
        """
        Remove the Repository specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, REPOSITORY_KEY)

    def remove_note(self, handle, transaction):
        """
        Remove the Note specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, NOTE_KEY)

    def remove_tag(self, handle, transaction):
        """
        Remove the Tag specified by the database handle from the
        database, preserving the change in the passed transaction.
        """
        self._do_remove(handle, transaction, TAG_KEY)

    ################################################################
    #
    # get_*_types methods
    #
    ################################################################

    def get_event_attribute_types(self):
        """
        Return a list of all Attribute types assocated with Event instances
        in the database.
        """
        return list(self.event_attributes)

    def get_event_types(self):
        """
        Return a list of all event types in the database.
        """
        return list(self.event_names)

    def get_person_event_types(self):
        """
        Deprecated:  Use get_event_types
        """
        return list(self.event_names)

    def get_person_attribute_types(self):
        """
        Return a list of all Attribute types assocated with Person instances
        in the database.
        """
        return list(self.individual_attributes)

    def get_family_attribute_types(self):
        """
        Return a list of all Attribute types assocated with Family instances
        in the database.
        """
        return list(self.family_attributes)

    def get_family_event_types(self):
        """
        Deprecated:  Use get_event_types
        """
        return list(self.event_names)

    def get_media_attribute_types(self):
        """
        Return a list of all Attribute types assocated with Media and MediaRef
        instances in the database.
        """
        return list(self.media_attributes)

    def get_family_relation_types(self):
        """
        Return a list of all relationship types assocated with Family
        instances in the database.
        """
        return list(self.family_rel_types)

    def get_child_reference_types(self):
        """
        Return a list of all child reference types assocated with Family
        instances in the database.
        """
        return list(self.child_ref_types)

    def get_event_roles(self):
        """
        Return a list of all custom event role names assocated with Event
        instances in the database.
        """
        return list(self.event_role_names)

    def get_name_types(self):
        """
        Return a list of all custom names types assocated with Person
        instances in the database.
        """
        return list(self.name_types)

    def get_origin_types(self):
        """
        Return a list of all custom origin types assocated with Person/Surname
        instances in the database.
        """
        return list(self.origin_types)

    def get_repository_types(self):
        """
        Return a list of all custom repository types assocated with Repository
        instances in the database.
        """
        return list(self.repository_types)

    def get_note_types(self):
        """
        Return a list of all custom note types assocated with Note instances
        in the database.
        """
        return list(self.note_types)

    def get_source_attribute_types(self):
        """
        Return a list of all Attribute types assocated with Source/Citation
        instances in the database.
        """
        return list(self.source_attributes)

    def get_source_media_types(self):
        """
        Return a list of all custom source media types assocated with Source
        instances in the database.
        """
        return list(self.source_media_types)

    def get_url_types(self):
        """
        Return a list of all custom names types assocated with Url instances
        in the database.
        """
        return list(self.url_types)

    def get_place_types(self):
        """
        Return a list of all custom place types assocated with Place instances
        in the database.
        """
        return list(self.place_types)

    ################################################################
    #
    # get_*_bookmarks methods
    #
    ################################################################

    def get_bookmarks(self):
        return self.bookmarks

    def get_citation_bookmarks(self):
        return self.citation_bookmarks

    def get_event_bookmarks(self):
        return self.event_bookmarks

    def get_family_bookmarks(self):
        return self.family_bookmarks

    def get_media_bookmarks(self):
        return self.media_bookmarks

    def get_note_bookmarks(self):
        return self.note_bookmarks

    def get_place_bookmarks(self):
        return self.place_bookmarks

    def get_repo_bookmarks(self):
        return self.repo_bookmarks

    def get_source_bookmarks(self):
        return self.source_bookmarks

    ################################################################
    #
    # Other methods
    #
    ################################################################

    def get_default_handle(self):
        return self._get_metadata("default-person-handle", None)

    def get_default_person(self):
        handle = self.get_default_handle()
        if handle:
            return self.get_person_from_handle(handle)
        return None

    def set_default_person_handle(self, handle):
        self._set_metadata("default-person-handle", handle)
        self.emit("home-person-changed")

    def get_mediapath(self):
        return self._get_metadata("media-path", None)

    def set_mediapath(self, mediapath):
        return self._set_metadata("media-path", mediapath)

    def get_surname_list(self):
        """
        Return the list of locale-sorted surnames contained in the database.
        """
        return self.surname_list

    def add_to_surname_list(self, person, batch_transaction):
        """
        Add surname to surname list
        """
        if batch_transaction:
            return
        name = None
        primary_name = person.get_primary_name()
        if primary_name:
            surname_list = primary_name.get_surname_list()
            if len(surname_list) > 0:
                name = surname_list[0].surname
        if name is None:
            return
        i = bisect.bisect(self.surname_list, name)
        if 0 < i <= len(self.surname_list):
            if self.surname_list[i - 1] != name:
                self.surname_list.insert(i, name)
        else:
            self.surname_list.insert(i, name)

    def remove_from_surname_list(self, person):
        """
        Check whether there are persons with the same surname left in
        the database.

        If not then we need to remove the name from the list.
        The function must be overridden in the derived class.
        """
        name = None
        primary_name = person.get_primary_name()
        if primary_name:
            surname_list = primary_name.get_surname_list()
            if len(surname_list) > 0:
                name = surname_list[0].surname
        if name is None:
            return
        if name in self.surname_list:
            self.surname_list.remove(name)

    def get_gender_stats(self):
        """
        Returns a dictionary of
        {given_name: (male_count, female_count, unknown_count)}
        """
        raise NotImplementedError

    def save_gender_stats(self, gstats):
        raise NotImplementedError

    def get_researcher(self):
        return self.owner

    def set_researcher(self, owner):
        self.owner.set_from(owner)

    def request_rebuild(self):
        self.emit("person-rebuild")
        self.emit("family-rebuild")
        self.emit("place-rebuild")
        self.emit("source-rebuild")
        self.emit("citation-rebuild")
        self.emit("media-rebuild")
        self.emit("event-rebuild")
        self.emit("repository-rebuild")
        self.emit("note-rebuild")
        self.emit("tag-rebuild")

    def get_save_path(self):
        return self._directory

    def _set_save_path(self, directory):
        self._directory = directory
        if directory:
            self.full_name = os.path.abspath(self._directory)
            self.path = self.full_name
            self.brief_name = os.path.basename(self._directory)
        else:
            self.full_name = None
            self.path = None
            self.brief_name = None

    def report_bm_change(self):
        """
        Add 1 to the number of bookmark changes during this session.
        """
        self._bm_changes += 1

    def db_has_bm_changes(self):
        """
        Return whethere there were bookmark changes during the session.
        """
        return self._bm_changes > 0

    def get_undodb(self):
        return self.undodb

    def undo(self, update_history=True):
        return self.undodb.undo(update_history)

    def redo(self, update_history=True):
        return self.undodb.redo(update_history)

    def get_summary(self):
        """
        Returns dictionary of summary item.
        Should include, if possible:

        _("Number of people")
        _("Version")
        _("Data version")
        """
        return {
            _("Number of people"): self.get_number_of_people(),
            _("Number of families"): self.get_number_of_families(),
            _("Number of sources"): self.get_number_of_sources(),
            _("Number of citations"): self.get_number_of_citations(),
            _("Number of events"): self.get_number_of_events(),
            _("Number of media"): self.get_number_of_media(),
            _("Number of places"): self.get_number_of_places(),
            _("Number of repositories"): self.get_number_of_repositories(),
            _("Number of notes"): self.get_number_of_notes(),
            _("Number of tags"): self.get_number_of_tags(),
            _("Schema version"): ".".join([str(v) for v in self.VERSION]),
        }

    def _order_by_person_key(self, person):
        """
        All non pa/matronymic surnames are used in indexing.
        pa/matronymic not as they change for every generation!
        returns a byte string
        """
        order_by = ""
        if person.primary_name:
            order_by_list = [
                surname.surname + " " + person.primary_name.first_name
                for surname in person.primary_name.surname_list
                if (
                    int(surname.origintype)
                    not in [NameOriginType.PATRONYMIC, NameOriginType.MATRONYMIC]
                )
            ]
            order_by = " ".join(order_by_list)
        return glocale.sort_key(order_by)

    def _get_person_data(self, person):
        """
        Given a Person, return primary_name.first_name and surname.
        """
        given_name = ""
        surname = ""
        if person:
            primary_name = person.get_primary_name()
            if primary_name:
                given_name = primary_name.get_first_name()
                surname_list = primary_name.get_surname_list()
                if len(surname_list) > 0:
                    surname_obj = surname_list[0]
                    if surname_obj:
                        surname = surname_obj.surname
        return (given_name, surname)

    def _get_place_data(self, place):
        """
        Given a Place, return the first PlaceRef handle.
        """
        enclosed_by = ""
        for placeref in place.get_placeref_list():
            enclosed_by = placeref.ref
            break
        return enclosed_by

    def _gramps_upgrade(self, version, directory, callback=None):
        """
        Here we do the calls for stepwise schema upgrades.
        We assume that we need to rebuild secondary and reference maps.
        """
        UpdateCallback.__init__(self, callback)

        start = time.time()

        from gramps.gen.db.upgrade import (
            gramps_upgrade_14,
            gramps_upgrade_15,
            gramps_upgrade_16,
            gramps_upgrade_17,
            gramps_upgrade_18,
            gramps_upgrade_19,
            gramps_upgrade_20,
            gramps_upgrade_21,
        )

        if version < 14:
            gramps_upgrade_14(self)
        if version < 15:
            gramps_upgrade_15(self)
        if version < 16:
            gramps_upgrade_16(self)
        if version < 17:
            gramps_upgrade_17(self)
        if version < 18:
            gramps_upgrade_18(self)
        if version < 19:
            gramps_upgrade_19(self)
        if version < 20:
            gramps_upgrade_20(self)
        if version < 21:
            gramps_upgrade_21(self)

        self.rebuild_secondary(callback)
        self.reindex_reference_map(callback)
        self.reset()

        self.set_schema_version(self.VERSION[0])
        LOG.debug("Upgrade time: %d seconds", int(time.time() - start))

    def get_schema_version(self):
        """Return current schema version as an int"""
        return int(self._get_metadata("version", default="0"))

    def set_schema_version(self, value):
        """set the current schema version"""
        self._set_metadata("version", str(value))

    def set_serializer(self, serializer_name):
        """
        Set the serializer to 'blob' or 'json'
        """
        if serializer_name == "blob":
            self.serializer = BlobSerializer
        elif serializer_name == "json":
            self.serializer = JSONSerializer


Database = DbGeneric
